import torch

from syft.generic import object_storage


def test_clear_objects():
    obj_storage = object_storage.ObjectStorage()

    x = torch.tensor(1)
    obj_storage.set_obj(x)

    objs = obj_storage.current_objects()

    assert len(objs) == 1
    assert objs[x.id] == x

    ret_val = obj_storage.clear_objects()

    objs = obj_storage.current_objects()
    assert len(objs) == 0
    assert ret_val == obj_storage


def test_clear_objects_return_None():
    obj_storage = object_storage.ObjectStorage()

    x = torch.tensor(1)
    obj_storage.set_obj(x)

    objs = obj_storage.current_objects()

    assert len(objs) == 1
    assert objs[x.id] == x

    ret_val = obj_storage.clear_objects(return_self=False)

    objs = obj_storage.current_objects()
    assert len(objs) == 0
    assert ret_val is None



# from syft.generic.tensor import AbstractTensor
# import torch
# import numpy as np
# from typing import Callable, List, Union
#
#
# class PolynomialTensor(AbstractTensor):
#     """
#     Tensor type to provide non-linear function approximations
#
#     MPC and Homomorphic Encryption are capable of performing some addition and logical operations.
#     Non-linear functions could be approximated as a series of approximated functions of basic arithmetic
#     operations using function approximations such as interpolation/Taylor series.
#
#     The polynomial tensor provides flexibility to consider every non-linear function as piecewise linear function
#     and fit over different intervals.
#     """
#
#     def __init__(self, function=lambda x: x, precision=10):
#         """
#         Args:
#             function[callable,Optional]: Function to applied to function approximation coefficients.
#                 Used to encrypt coefficients.
#             precision[integer]: Precision of approximated values
#         """
#
#         self.function = function
#         self.precision = precision
#
#         # Stores parameters of function approximations such as precision, degree, piecewise functions and base function
#         self.function_attr = {}
#
#         # Stores fitted function
#         self.func_approx = {}
#
#         self.default_functions()
#
#     def default_functions(self):
#         """Initializes default function approximations exp, log, sigmoid and tanh"""
#
#         self.add_function(
#             "exp",
#             10,
#             [[0, 10, 100, 10, self.fit_function], [-10, 0, 100, 10, self.fit_function]],
#             lambda x: np.exp(x),
#         )
#         self.add_function("log", 10, [[1, 10, 100, 10, self.fit_function]], lambda x: np.log(x))
#         self.add_function(
#             "sigmoid", 10, [[-10, 10, 100, 10, self.fit_function]], (lambda x: 1 / (1 + np.exp(-x)))
#         )
#         self.add_function(
#             "tanh",
#             10,
#             [[0, 10, 1000, 10, self.fit_function], [-10, 0, 1000, 10, self.fit_function]],
#             lambda x: np.tanh(x),
#         )
#
#     def add_function(self, name, degree, piecewise, function):
#         """Add function to function_attr dictionary.
#
#         Args:
#             name[str]: Name of function
#             degree[int]: Degree of function
#             piecewise[List]: List of piecewise functions in format [min_val of fit,max_val of fit,step of fit,function to fit values]
#             function[callable]: Base function
#         """
#
#         self.function_attr[name + "_degree"] = degree
#         self.function_attr[name + "_piecewise"] = piecewise
#         self.function_attr[name + "_function"] = function
#
#         self.func_approx[name] = self.piecewise_linear_fit(
#             name, self.function_attr[name + "_piecewise"]
#         )
#
#     def get_val(self, name, x):
#         """Get value of given function approximation
#
#         Args:
#             name[str]: Name of function
#             value[torch tensor,float,integer]: Value to be approximated
#
#         Returns:
#             Approximated value using given function approximation
#         """
#
#         value = x
#
#         if type(x) == torch.Tensor:
#
#             return value.apply_(lambda k: self.piecewise_linear_eval(self.func_approx[name], k))
#
#         return self.piecewise_linear_eval(self.func_approx[name], x)
#
#     def interpolate(
#         self, function: Callable, interval: List[Union[int, float]], degree: int = 10
#     ) -> np.poly1d:
#
#         """Returns a interpolated version of given function using Numpy's polyfit method
#
#         Args:
#             function (a lambda function): Base function to be approximated
#             interval (list of floats/integers): Interval of values to be approximated
#             degree (Integer): Degree of polynomial approximation
#             precision (Integer): Precision of coefficients
#
#         Returns:
#             f_interpolated (Numpy poly1d): Approximated Function
#         """
#
#         # function we wish to approximate
#         f_real = function
#         # interval over which we wish to optimize
#         f_interval = interval
#
#         # interpolate polynomial of given max degree
#         degree = 10
#         coefs = np.polyfit(f_interval, f_real(f_interval), degree)
#
#         # reduce precision of interpolated coefficients
#         precision = self.precision
#         coefs = [int(x * 10 ** precision) / 10 ** precision for x in coefs]
#
#         # approximation function
#         f_interpolated = np.poly1d(coefs)
#
#         return f_interpolated
#
#     def apply_coefs(self, polyinstance, function):
#         """Apply a given function over Numpy interpolation instances.This function could be used
#         to encrypt coefficients of function approximations approximated using interpolation
#
#         Args:
#             polyinstance (Numpy poly1d): Interpolation instance
#             function (Callable): Function to be applied
#         """
#
#         val = torch.from_numpy(polyinstance.coef)
#         return function(val)
#
#     def piecewise_linear_fit(self, name, array):
#         """Fit a piecewise linear function. This can be used to approximate a non-linear function
#         as separate linear functions valid for separate ranges.
#         For instance function approximations are more accurate for exponential when separate instances
#         of interpolation are fit between -10 to 0 and 0 to 10.
#
#         Args:
#             array[2D List]: Each instance of list must take four values [min_val, steps, max_val,
#                 function approximation method]
#
#         Returns:
#             array[2D List]: Each instance of list with four values [min_val,max_val,Approximated function]
#         """
#
#         arguments = []
#
#         for element in array:
#
#             min_val = element[0]
#             max_val = element[1]
#             steps = element[2]
#             degree = element[3]
#             function = element[4]
#             arguments.append(
#                 [
#                     min_val,
#                     max_val,
#                     function(name, min_val=min_val, max_val=max_val, steps=steps, degree=degree),
#                 ]
#             )
#
#         return arguments
#
#     def piecewise_linear_eval(self, data, x):
#         """Get approximated value for a given function. This takes only scalar value.
#         If you have a Numpy array or torch tensor consider passing it using a lambda
#         or torch.apply_ method.
#
#         Args:
#             data[2D List]: Instance of piecewise linear fit taking values [min_val, max_val,
#                 function approximation method]
#             x[Float or Integer]: Value to be approximated
#         """
#
#         for element in data:
#
#             min_val = element[0]
#             max_val = element[1]
#             function = element[2]
#
#             if min_val <= x <= max_val:
#
#                 return function(x)
#
#     def fit_function(self, name, min_val=0, max_val=10, steps=100, degree=10) -> np.poly1d:
#         """Interpolated approximation of given function
#
#         Args:
#             name: Name of function as defined in self.setting
#             min_val: Minimum range of interpolation fit
#             max_val: Maximum range of interpolation fit
#             steps:   Steps of interpolation fit
#             degree: Degree of interpolation fit
#             function: The function used to encrypt function approximation coefficients
#
#         Returns:
#             f_interpolated (Numpy Poly1d): Approximated function
#         """
#
#         fitted_function = self.interpolate(
#             self.function_attr[name + "_" + "function"],
#             np.linspace(min_val, max_val, steps),
#             degree=degree,
#         )
#
#         fitted_function = self.apply_coefs(fitted_function, self.function)
#
#         return np.poly1d(fitted_function)
#
#     def sigmoid(self, x):
#         """Method provides Sigmoid function approximation interms of Taylor Series
#
#         Args:
#             x: Torch tensor
#
#         Returns:
#             approximation of the sigmoid function as a torch tensor
#         """
#
#         return (
#             (self.function(1 / 2))
#             + ((x) * self.function(1 / 4))
#             - ((x ** 3) * self.function(1 / 48))
#             + ((x ** 5) * self.function((1 / 480)))
#         )
#
#     def exp(self, x):
#         """
#         Method provides exponential function approximation interms of Taylor Series
#
#         Args:
#             x: Torch tensor
#
#         Returns:
#             approximation of the sigmoid function as a torch tensor
#         """
#
#         return (
#             self.function(1)
#             + self.function(x)
#             + (x ** 2) * (self.function(1 / 2))
#             + (x ** 3) * (self.function(1 / 6))
#             + (x ** 4) * (self.function(1 / (24)))
#             + (x ** 5) * (self.function(1 / (120)))
#             + (x ** 6) * (self.function(1 / (840)))
#             + (x ** 7) * (self.function(1 / (6720)))
#         )



from distutils.version import LooseVersion
import importlib
import logging

logger = logging.getLogger(__name__)

try:
    import tensorflow

    if LooseVersion(tensorflow.__version__) < LooseVersion("2.0.0"):
        raise ImportError()
    pstf_spec = importlib.util.find_spec("syft_tensorflow")
    tensorflow_available = pstf_spec is not None
except ImportError:
    tensorflow_available = False


tfe_spec = importlib.util.find_spec("tf_encrypted")
tfe_available = tfe_spec is not None


torch_spec = importlib.util.find_spec("torch")
torch_available = torch_spec is not None



import torch as th
import syft as sy

from syft.frameworks.torch.federated import utils
from syft.frameworks.torch import federated


def test_extract_batches_per_worker(workers):
    bob = workers["bob"]
    alice = workers["alice"]

    datasets = [
        federated.BaseDataset(th.tensor([1, 2]), th.tensor([1, 2])).send(bob),
        federated.BaseDataset(th.tensor([3, 4, 5, 6]), th.tensor([3, 4, 5, 6])).send(alice),
    ]
    fed_dataset = sy.FederatedDataset(datasets)

    fdataloader = sy.FederatedDataLoader(fed_dataset, batch_size=2, shuffle=True)

    batches = utils.extract_batches_per_worker(fdataloader)

    assert len(batches.keys()) == len(
        datasets
    ), "each worker should appear as key in the batches dictionary"


def test_add_model():
    class Net(th.nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.fc1 = th.nn.Linear(2, 2)

    weight1 = th.tensor([1.0, 2.0, 3.0, 4.0])
    weight2 = th.tensor([11.0, 22.0, 33.0, 44.0])

    bias1 = th.tensor([-1.0, -2.0])
    bias2 = th.tensor([1.0, 2.0])

    net1 = Net()
    params1 = net1.named_parameters()
    dict_params1 = dict(params1)
    with th.no_grad():
        dict_params1["fc1.weight"].set_(weight1)
        dict_params1["fc1.bias"].set_(bias1)

    net2 = Net()
    params2 = net2.named_parameters()
    dict_params2 = dict(params2)
    with th.no_grad():
        dict_params2["fc1.weight"].set_(weight2)
        dict_params2["fc1.bias"].set_(bias2)

    new_model = utils.add_model(net1, net2)

    assert (new_model.fc1.weight.data == (weight1 + weight2)).all()
    assert (new_model.fc1.bias.data == (bias1 + bias2)).all()


def test_scale_model():
    class Net(th.nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.fc1 = th.nn.Linear(2, 2)

    weight1 = th.tensor([1.0, 2.0, 3.0, 4.0])

    bias1 = th.tensor([-1.0, -2.0])

    net1 = Net()
    params1 = net1.named_parameters()
    dict_params1 = dict(params1)
    with th.no_grad():
        dict_params1["fc1.weight"].set_(weight1)
        dict_params1["fc1.bias"].set_(bias1)

    scale = 2.0

    new_model = utils.scale_model(net1, scale)

    assert (new_model.fc1.weight.data == (weight1 * scale)).all()
    assert (new_model.fc1.bias.data == (bias1 * scale)).all()


def test_accuracy():
    pred = th.tensor([[0.95, 0.02, 0.03], [0.3, 0.4, 0.3], [0.0, 0.0, 1.0]])

    target = th.tensor([0.0, 1.0, 2.0])

    acc = utils.accuracy(pred, target)

    assert acc == 1.0

    target = th.tensor([2.0, 0.0, 2.0])

    acc = utils.accuracy(pred, target)

    assert acc == 1.0 / 3.0



import torch

import syft
from syft.workers.abstract import AbstractWorker
from syft.generic.frameworks.hook import hook_args
from syft.generic.pointers.multi_pointer import MultiPointerTensor
from syft.generic.tensor import AbstractTensor
from syft.frameworks.torch.tensors.interpreters.additive_shared import AdditiveSharingTensor
from syft.generic.frameworks.overload import overloaded


class FixedPrecisionTensor(AbstractTensor):
    def __init__(
        self,
        owner=None,
        id=None,
        field: int = 2 ** 62,
        base: int = 10,
        precision_fractional: int = 3,
        kappa: int = 1,
        tags: set = None,
        description: str = None,
    ):
        """Initializes a Fixed Precision tensor, which encodes all decimal point
        values using an underlying integer value.

        The FixedPrecision enables to manipulate floats over an interface which
        supports only integers, Such as _SPDZTensor.

        This is done by specifying a precision p and given a float x,
        multiply it with 10**p before rounding to an integer (hence you keep
        p decimals)

        Args:
            owner: An optional BaseWorker object to specify the worker on which
                the tensor is located.
            id: An optional string or integer id of the FixedPrecisionTensor.
        """
        super().__init__(tags, description)

        self.owner = owner
        self.id = id if id else syft.ID_PROVIDER.pop()
        self.child = None

        self.field = field
        self.base = base
        self.precision_fractional = precision_fractional
        self.kappa = kappa

    def get_class_attributes(self):
        """
        Specify all the attributes need to build a wrapper correctly when returning a response,
        for example precision_fractional is important when wrapping the result of a method
        on a self which is a fixed precision tensor with a non default precision_fractional.
        """
        return {
            "field": self.field,
            "base": self.base,
            "precision_fractional": self.precision_fractional,
            "kappa": self.kappa,
        }

    @property
    def data(self):
        return self

    @data.setter
    def data(self, new_data):
        self.child = new_data.child
        return self

    @property
    def grad(self):
        """
        Gradient makes no sense for Fixed Precision Tensor, so we make it clear
        that if someone query .grad on a Fixed Precision Tensor it doesn't error
        but returns grad and can't be set
        """
        return None

    def backward(self, *args, **kwargs):
        """Calling backward on Precision Tensor doesn't make sense, but sometimes a call
        can be propagated downward the chain to an Precision Tensor (for example in
        create_grad_objects), so we just ignore the call."""
        pass

    def attr(self, attr_name):
        return self.__getattribute__(attr_name)

    def fix_precision(self, check_range=True):
        """This method encodes the .child object using fixed precision"""

        rational = self.child

        upscaled = (rational * self.base ** self.precision_fractional).long()

        if check_range:
            assert (
                upscaled.abs() < (self.field / 2)
            ).all(), (
                f"{rational} cannot be correctly embedded: choose bigger field or a lower precision"
            )

        field_element = upscaled % self.field
        field_element.owner = rational.owner

        self.child = field_element
        return self

    def float_precision(self):
        """this method returns a new tensor which has the same values as this
        one, encoded with floating point precision"""

        value = self.child.long() % self.field
        torch_max_value = torch.tensor(self.field).long()

        gate = value.native_gt(torch_max_value / 2).long()
        neg_nums = (value - self.field) * gate
        pos_nums = value * (1 - gate)
        result = (neg_nums + pos_nums).float() / (self.base ** self.precision_fractional)

        return result

    def truncate(self, precision_fractional, check_sign=True):
        truncation = self.base ** precision_fractional

        # We need to make sure that values are truncated "towards 0"
        # i.e. for a field of 100, 70 (equivalent to -30), should be truncated
        # at 97 (equivalent to -3), not 7
        if isinstance(self.child, AdditiveSharingTensor) or not check_sign:  # Handle FPT>(wrap)>AST
            self.child = self.child / truncation
            return self
        else:
            torch_max_value = torch.tensor(self.field).long()
            gate = self.child.native_gt(torch_max_value / 2).long()
            neg_nums = (self.child - self.field) / truncation + self.field
            pos_nums = self.child / truncation
            self.child = neg_nums * gate + pos_nums * (1 - gate)
            return self

    @overloaded.method
    def add(self, _self, other):
        """Add two fixed precision tensors together.
        """
        if isinstance(other, int):
            scaled_int = other * self.base ** self.precision_fractional
            return getattr(_self, "add")(scaled_int)

        if isinstance(_self, AdditiveSharingTensor) and isinstance(other, torch.Tensor):
            # If we try to add a FPT>(wrap)>AST and a FPT>torch.tensor,
            # we want to perform AST + torch.tensor
            other = other.wrap()
        elif isinstance(other, AdditiveSharingTensor) and isinstance(_self, torch.Tensor):
            # If we try to add a FPT>torch.tensor and a FPT>(wrap)>AST,
            # we swap operators so that we do the same operation as above
            _self, other = other, _self.wrap()

        response = getattr(_self, "add")(other)
        response %= self.field  # Wrap around the field

        return response

    __add__ = add

    def add_(self, value_or_tensor, tensor=None):
        if tensor is None:
            result = self.add(value_or_tensor)
        else:
            result = self.add(value_or_tensor * tensor)

        self.child = result.child
        return self

    def __iadd__(self, other):
        """Add two fixed precision tensors together.
        """
        self.child = self.add(other).child

        return self

    @overloaded.method
    def sub(self, _self, other):
        """Subtracts a fixed precision tensor from another one.
        """
        if isinstance(other, int):
            scaled_int = other * self.base ** self.precision_fractional
            return getattr(_self, "sub")(scaled_int)

        if isinstance(_self, AdditiveSharingTensor) and isinstance(other, torch.Tensor):
            # If we try to subtract a FPT>(wrap)>AST and a FPT>torch.tensor,
            # we want to perform AST - torch.tensor
            other = other.wrap()
        elif isinstance(other, AdditiveSharingTensor) and isinstance(_self, torch.Tensor):
            # If we try to subtract a FPT>torch.tensor and a FPT>(wrap)>AST,
            # we swap operators so that we do the same operation as above
            _self, other = -other, -_self.wrap()

        response = getattr(_self, "sub")(other)
        response %= self.field  # Wrap around the field

        return response

    __sub__ = sub

    def sub_(self, value_or_tensor, tensor=None):
        if tensor is None:
            result = self.sub(value_or_tensor)
        else:
            result = self.sub(value_or_tensor * tensor)

        self.child = result.child
        return self

    def __isub__(self, other):
        self.child = self.sub(other).child

        return self

    @overloaded.method
    def t(self, _self, *args, **kwargs):
        """Transpose a tensor. Hooked is handled by the decorator"""
        response = getattr(_self, "t")(*args, **kwargs)

        return response

    def mul_and_div(self, other, cmd):
        """
        Hook manually mul and div to add the trucation/rescaling part
        which is inherent to these operations in the fixed precision setting
        """
        changed_sign = False
        if isinstance(other, FixedPrecisionTensor):
            assert (
                self.precision_fractional == other.precision_fractional
            ), "In mul and div, all args should have the same precision_fractional"
            assert self.base == other.base, "In mul and div, all args should have the same base"

        if isinstance(other, (int, torch.Tensor, AdditiveSharingTensor)):
            new_self = self.child
            new_other = other

        elif isinstance(self.child, (AdditiveSharingTensor, MultiPointerTensor)) and isinstance(
            other.child, torch.Tensor
        ):
            # If operands are FPT>AST and FPT>torch.tensor,
            # we want to perform the operation on AST and torch.tensor
            if cmd == "mul":
                new_self = self.child
            elif cmd == "div":
                new_self = self.child * self.base ** self.precision_fractional
            new_other = other

        elif isinstance(other.child, (AdditiveSharingTensor, MultiPointerTensor)) and isinstance(
            self.child, torch.Tensor
        ):
            # If operands are FPT>torch.tensor and FPT>AST,
            # we swap operators so that we do the same operation as above
            if cmd == "mul":
                new_self = other.child
                new_other = self
            elif cmd == "div":
                # TODO how to divide by AST?
                raise NotImplementedError(
                    "Division of a FixedPrecisionTensor by an AdditiveSharingTensor not implemented"
                )

        elif (
            cmd == "mul"
            and isinstance(self.child, (AdditiveSharingTensor, MultiPointerTensor))
            and isinstance(other.child, (AdditiveSharingTensor, MultiPointerTensor))
        ):
            # If we try to multiply a FPT>torch.tensor with a FPT>AST,
            # we swap operators so that we do the same operation as above
            new_self, new_other, _ = hook_args.unwrap_args_from_method("mul", self, other, None)

        else:
            # Replace all syft tensor with their child attribute
            new_self, new_other, _ = hook_args.unwrap_args_from_method(cmd, self, other, None)

            # To avoid problems with negative numbers
            # we take absolute value of the operands
            # The problems could be 1) bad truncation for multiplication
            # 2) overflow when scaling self in division

            # sgn_self is 1 when new_self is positive else it's 0
            # The comparison is different is new_self is a torch tensor or an AST
            sgn_self = (
                (new_self < self.field // 2).long()
                if isinstance(new_self, torch.Tensor)
                else new_self > 0
            )
            pos_self = new_self * sgn_self
            neg_self = (
                (self.field - new_self) * (1 - sgn_self)
                if isinstance(new_self, torch.Tensor)
                else new_self * (sgn_self - 1)
            )
            new_self = neg_self + pos_self

            # sgn_other is 1 when new_other is positive else it's 0
            # The comparison is different is new_other is a torch tensor or an AST
            sgn_other = (
                (new_other < self.field // 2).long()
                if isinstance(new_other, torch.Tensor)
                else new_other > 0
            )
            pos_other = new_other * sgn_other
            neg_other = (
                (self.field - new_other) * (1 - sgn_other)
                if isinstance(new_other, torch.Tensor)
                else new_other * (sgn_other - 1)
            )
            new_other = neg_other + pos_other

            # If both have the same sign, sgn is 1 else it's 0
            # To be able to write sgn = 1 - (sgn_self - sgn_other) ** 2,
            # we would need to overload the __add__ for operators int and AST.
            sgn = -(sgn_self - sgn_other) ** 2 + 1
            changed_sign = True

            if cmd == "div":
                new_self *= self.base ** self.precision_fractional

        # Send it to the appropriate class and get the response
        response = getattr(new_self, cmd)(new_other)

        # Put back SyftTensor on the tensors found in the response
        response = hook_args.hook_response(
            cmd, response, wrap_type=type(self), wrap_args=self.get_class_attributes()
        )

        if not isinstance(other, (int, torch.Tensor, AdditiveSharingTensor)):
            if cmd == "mul":
                # If operation is mul, we need to truncate
                response = response.truncate(self.precision_fractional, check_sign=False)

            response %= self.field  # Wrap around the field

            if changed_sign:
                # Give back its sign to response
                pos_res = response * sgn
                neg_res = response * (sgn - 1)
                response = neg_res + pos_res

        else:
            response %= self.field  # Wrap around the field

        return response

    def mul(self, other):
        return self.mul_and_div(other, "mul")

    __mul__ = mul

    def __imul__(self, other):
        self = self.mul_and_div(other, "mul")
        return self

    mul_ = __imul__

    def div(self, other):
        return self.mul_and_div(other, "div")

    __truediv__ = div

    def __itruediv__(self, other):
        self = self.mul_and_div(other, "div")
        return self

    def pow(self, power):
        """
        Compute integer power of a number by recursion using mul

        This uses the following trick:
         - Divide power by 2 and multiply base to itself (if the power is even)
         - Decrement power by 1 to make it even and then follow the first step
        """
        base = self

        result = 1
        while power > 0:
            # If power is odd
            if power % 2 == 1:
                result = result * base

            # Divide the power by 2
            power = power // 2
            # Multiply base to itself
            base = base * base

        return result

    __pow__ = pow

    def matmul(self, *args, **kwargs):
        """
        Hook manually matmul to add the truncation part which is inherent to multiplication
        in the fixed precision setting
        """

        other = args[0]

        if isinstance(other, FixedPrecisionTensor):
            assert (
                self.precision_fractional == other.precision_fractional
            ), "In matmul, all args should have the same precision_fractional"

        if isinstance(self.child, AdditiveSharingTensor) and isinstance(other.child, torch.Tensor):
            # If we try to matmul a FPT>(wrap)>AST with a FPT>torch.tensor,
            # we want to perform AST @ torch.tensor
            new_self = self.child
            new_args = (other,)
            new_kwargs = kwargs

        elif isinstance(other.child, AdditiveSharingTensor) and isinstance(
            self.child, torch.Tensor
        ):
            # If we try to matmul a FPT>torch.tensor with a FPT>(wrap)>AST,
            # we swap operators so that we do the same operation as above
            new_self = other.child
            new_args = (self,)
            new_kwargs = kwargs

        else:
            # Replace all syft tensor with their child attribute
            new_self, new_args, new_kwargs = hook_args.unwrap_args_from_method(
                "matmul", self, args, kwargs
            )

        # Send it to the appropriate class and get the response
        response = getattr(new_self, "matmul")(*new_args, **new_kwargs)

        # Put back SyftTensor on the tensors found in the response
        response = hook_args.hook_response(
            "matmul", response, wrap_type=type(self), wrap_args=self.get_class_attributes()
        )

        response %= self.field  # Wrap around the field
        response = response.truncate(other.precision_fractional)

        return response

    __matmul__ = matmul
    mm = matmul

    @overloaded.method
    def __gt__(self, _self, other):
        result = _self.__gt__(other)
        return result.long() * self.base ** self.precision_fractional

    @overloaded.method
    def __ge__(self, _self, other):
        result = _self.__ge__(other)
        return result.long() * self.base ** self.precision_fractional

    @overloaded.method
    def __lt__(self, _self, other):
        result = _self.__lt__(other)
        return result.long() * self.base ** self.precision_fractional

    @overloaded.method
    def __le__(self, _self, other):
        result = _self.__le__(other)
        return result.long() * self.base ** self.precision_fractional

    @overloaded.method
    def eq(self, _self, other):
        result = _self.eq(other)
        return result.long() * self.base ** self.precision_fractional

    __eq__ = eq

    @staticmethod
    @overloaded.module
    def torch(module):
        def add(self, other):
            return self.__add__(other)

        module.add = add

        def sub(self, other):
            return self.__sub__(other)

        module.sub = sub

        def mul(self, other):
            return self.__mul__(other)

        module.mul = mul

        def div(self, other):
            return self.__truediv__(other)

        module.div = div

        def matmul(self, other):
            return self.matmul(other)

        module.matmul = matmul
        module.mm = matmul

        def addmm(bias, input_tensor, weight):
            matmul = input_tensor.matmul(weight)
            result = bias.add(matmul)
            return result

        module.addmm = addmm

        def sigmoid(tensor):
            """
            Overloads torch.sigmoid to be able to use MPC
            Approximation with polynomial interpolation of degree 5 over [-8,8]
            Ref: https://mortendahl.github.io/2017/04/17/private-deep-learning-with-mpc/#approximating-sigmoid
            """

            weights = [0.5, 1.91204779e-01, -4.58667307e-03, 4.20690803e-05]
            degrees = [0, 1, 3, 5]

            max_degree = degrees[-1]
            max_idx = degrees.index(max_degree)

            # initiate with term of degree 0 to avoid errors with tensor ** 0
            result = (tensor * 0 + 1) * torch.tensor(weights[0]).fix_precision().child
            for w, d in zip(weights[1:max_idx], degrees[1:max_idx]):
                result += (tensor ** d) * torch.tensor(w).fix_precision().child

            return result

        module.sigmoid = sigmoid

        def tanh(tensor):
            """
            Overloads torch.tanh to be able to use MPC
            """

            result = 2 * sigmoid(2 * tensor) - 1

            return result

        module.tanh = tanh

        def dot(self, other):
            return self.__mul__(other).sum()

        module.dot = dot

        def conv2d(
            input,
            weight,
            bias=None,
            stride=1,
            padding=0,
            dilation=1,
            groups=1,
            padding_mode="zeros",
        ):
            """
            Overloads torch.conv2d to be able to use MPC on convolutional networks.
            The idea is to build new tensors from input and weight to compute a
            matrix multiplication equivalent to the convolution.

            Args:
                input: input image
                weight: convolution kernels
                bias: optional additive bias
                stride: stride of the convolution kernels
                padding:
                dilation: spacing between kernel elements
                groups:
                padding_mode: type of padding, should be either 'zeros' or 'circular' but 'reflect' and 'replicate' accepted
            Returns:
                the result of the convolution as an AdditiveSharingTensor
            """
            # Currently, kwargs are not unwrapped by hook_args
            # So this needs to be done manually
            if bias.is_wrapper:
                bias = bias.child

            assert len(input.shape) == 4
            assert len(weight.shape) == 4

            # Change to tuple if not one
            stride = torch.nn.modules.utils._pair(stride)
            padding = torch.nn.modules.utils._pair(padding)
            dilation = torch.nn.modules.utils._pair(dilation)

            # Extract a few useful values
            batch_size, nb_channels_in, nb_rows_in, nb_cols_in = input.shape
            nb_channels_out, nb_channels_kernel, nb_rows_kernel, nb_cols_kernel = weight.shape

            if bias is not None:
                assert len(bias) == nb_channels_out

            # Check if inputs are coherent
            assert nb_channels_in == nb_channels_kernel * groups
            assert nb_channels_in % groups == 0
            assert nb_channels_out % groups == 0

            # Compute output shape
            nb_rows_out = int(
                ((nb_rows_in + 2 * padding[0] - dilation[0] * (nb_rows_kernel - 1) - 1) / stride[0])
                + 1
            )
            nb_cols_out = int(
                ((nb_cols_in + 2 * padding[1] - dilation[1] * (nb_cols_kernel - 1) - 1) / stride[1])
                + 1
            )

            # Apply padding to the input
            if padding != (0, 0):
                padding_mode = "constant" if padding_mode == "zeros" else padding_mode
                input = torch.nn.functional.pad(
                    input, (padding[1], padding[1], padding[0], padding[0]), padding_mode
                )
                # Update shape after padding
                nb_rows_in += 2 * padding[0]
                nb_cols_in += 2 * padding[1]

            # We want to get relative positions of values in the input tensor that are used by one filter convolution.
            # It basically is the position of the values used for the top left convolution.
            pattern_ind = []
            for ch in range(nb_channels_in):
                for r in range(nb_rows_kernel):
                    for c in range(nb_cols_kernel):
                        pixel = r * nb_cols_in * dilation[0] + c * dilation[1]
                        pattern_ind.append(pixel + ch * nb_rows_in * nb_cols_in)

            # The image tensor is reshaped for the matrix multiplication:
            # on each row of the new tensor will be the input values used for each filter convolution
            # We will get a matrix [[in values to compute out value 0],
            #                       [in values to compute out value 1],
            #                       ...
            #                       [in values to compute out value nb_rows_out*nb_cols_out]]
            im_flat = input.view(batch_size, -1)
            im_reshaped = []
            for cur_row_out in range(nb_rows_out):
                for cur_col_out in range(nb_cols_out):
                    # For each new output value, we just need to shift the receptive field
                    offset = cur_row_out * stride[0] * nb_cols_in + cur_col_out * stride[1]
                    tmp = [ind + offset for ind in pattern_ind]
                    im_reshaped.append(im_flat[:, tmp])
            im_reshaped = torch.stack(im_reshaped).permute(1, 0, 2)

            # The convolution kernels are also reshaped for the matrix multiplication
            # We will get a matrix [[weights for out channel 0],
            #                       [weights for out channel 1],
            #                       ...
            #                       [weights for out channel nb_channels_out]].TRANSPOSE()
            weight_reshaped = weight.view(nb_channels_out // groups, -1).t()

            # Now that everything is set up, we can compute the result
            if groups > 1:
                res = []
                chunks_im = torch.chunk(im_reshaped, groups, dim=2)
                chunks_weights = torch.chunk(weight_reshaped, groups, dim=0)
                for g in range(groups):
                    tmp = chunks_im[g].matmul(chunks_weights[g])
                    res.append(tmp)
                res = torch.cat(res, dim=2)
            else:
                res = im_reshaped.matmul(weight_reshaped)

            # Add a bias if needed
            if bias is not None:
                res += bias

            # ... And reshape it back to an image
            res = (
                res.permute(0, 2, 1)
                .view(batch_size, nb_channels_out, nb_rows_out, nb_cols_out)
                .contiguous()
            )
            return res

        module.conv2d = conv2d

        # You can also overload functions in submodules!
        @overloaded.module
        def nn(module):
            """
            The syntax is the same, so @overloaded.module handles recursion
            Note that we don't need to add the @staticmethod decorator
            """

            @overloaded.module
            def functional(module):
                def linear(*args):
                    """
                    Un-hook the function to have its detailed behaviour
                    """
                    return torch.nn.functional.native_linear(*args)

                module.linear = linear

            module.functional = functional

        # Modules should be registered just like functions
        module.nn = nn

    @classmethod
    def handle_func_command(cls, command):
        """
        Receive an instruction for a function to be applied on a FixedPrecision Tensor,
        Perform some specific action (like logging) which depends of the
        instruction content, replace in the args all the FPTensors with
        their child attribute, forward the command instruction to the
        handle_function_command of the type of the child attributes, get the
        response and replace a FixedPrecision on top of all tensors found in
        the response.
        :param command: instruction of a function command: (command name,
        <no self>, arguments[, kwargs])
        :return: the response of the function command
        """
        cmd, _, args, kwargs = command

        tensor = args[0] if not isinstance(args[0], (tuple, list)) else args[0][0]

        # Check that the function has not been overwritten
        try:
            # Try to get recursively the attributes in cmd = "<attr1>.<attr2>.<attr3>..."
            cmd = cls.rgetattr(cls, cmd)
            return cmd(*args, **kwargs)
        except AttributeError:
            pass

        # Replace all FixedPrecisionTensor with their child attribute
        new_args, new_kwargs, new_type = hook_args.unwrap_args_from_function(cmd, args, kwargs)

        # build the new command
        new_command = (cmd, None, new_args, new_kwargs)

        # Send it to the appropriate class and get the response
        response = new_type.handle_func_command(new_command)

        # Put back FixedPrecisionTensor on the tensors found in the response
        response = hook_args.hook_response(
            cmd, response, wrap_type=cls, wrap_args=tensor.get_class_attributes()
        )

        return response

    def share(self, *owners, field=None, crypto_provider=None):
        if field is None:
            field = self.field
        else:
            assert (
                field == self.field
            ), "When sharing a FixedPrecisionTensor, the field of the resulting AdditiveSharingTensor \
                must be the same as the one of the original tensor"
        self.child = self.child.share(
            *owners, field=field, crypto_provider=crypto_provider, no_wrap=True
        )
        return self

    @staticmethod
    def simplify(tensor: "FixedPrecisionTensor") -> tuple:
        """Takes the attributes of a FixedPrecisionTensor and saves them in a tuple.

        Args:
            tensor: a FixedPrecisionTensor.

        Returns:
            tuple: a tuple holding the unique attributes of the fixed precision tensor.
        """
        chain = None
        if hasattr(tensor, "child"):
            chain = syft.serde._simplify(tensor.child)

        return (
            syft.serde._simplify(tensor.id),
            tensor.field,
            tensor.base,
            tensor.precision_fractional,
            tensor.kappa,
            syft.serde._simplify(tensor.tags),
            syft.serde._simplify(tensor.description),
            chain,
        )

    @staticmethod
    def detail(worker: AbstractWorker, tensor_tuple: tuple) -> "FixedPrecisionTensor":
        """
            This function reconstructs a FixedPrecisionTensor given it's attributes in form of a tuple.
            Args:
                worker: the worker doing the deserialization
                tensor_tuple: a tuple holding the attributes of the FixedPrecisionTensor
            Returns:
                FixedPrecisionTensor: a FixedPrecisionTensor
            Examples:
                shared_tensor = detail(data)
            """

        tensor_id, field, base, precision_fractional, kappa, tags, description, chain = tensor_tuple

        tensor = FixedPrecisionTensor(
            owner=worker,
            id=syft.serde._detail(worker, tensor_id),
            field=field,
            base=base,
            precision_fractional=precision_fractional,
            kappa=kappa,
            tags=syft.serde._detail(worker, tags),
            description=syft.serde._detail(worker, description),
        )

        if chain is not None:
            chain = syft.serde._detail(worker, chain)
            tensor.child = chain

        return tensor


### Register the tensor with hook_args.py ###
hook_args.default_register_tensor(FixedPrecisionTensor)



import syft
from syft.generic.tensor import AbstractTensor
from syft.generic.frameworks.hook import hook_args
from syft.generic.frameworks.overload import overloaded
from syft.workers.abstract import AbstractWorker
import syft as sy


class LoggingTensor(AbstractTensor):
    def __init__(self, owner=None, id=None, tags=None, description=None):
        """Initializes a LoggingTensor, whose behaviour is to log all operations
        applied on it.

        Args:
            owner: An optional BaseWorker object to specify the worker on which
                the tensor is located.
            id: An optional string or integer id of the LoggingTensor.
        """
        super().__init__(id=id, owner=owner, tags=tags, description=description)

    # Method overloading

    @overloaded.method
    def add(self, _self, *args, **kwargs):
        """
        Here is an example of how to use the @overloaded.method decorator. To see
        what this decorator do, just look at the next method manual_add: it does
        exactly the same but without the decorator.

        Note the subtlety between self and _self: you should use _self and NOT self.
        """
        print("Log method add")
        response = getattr(_self, "add")(*args, **kwargs)

        return response

    def manual_add(self, *args, **kwargs):
        """
        Here is the version of the add method without the decorator: as you can see
        it is much more complicated. However you might need sometimes to specify
        some particular behaviour: so here what to start from :)
        """
        # Replace all syft tensor with their child attribute
        new_self, new_args, new_kwargs = hook_args.unwrap_args_from_method(
            "add", self, args, kwargs
        )

        print("Log method manual_add")
        # Send it to the appropriate class and get the response
        response = getattr(new_self, "add")(*new_args, **new_kwargs)

        # Put back SyftTensor on the tensors found in the response
        response = hook_args.hook_response("add", response, wrap_type=type(self))
        return response

    # Module & Function overloading

    # We overload two torch functions:
    # - torch.add
    # - torch.nn.functional.relu

    @staticmethod
    @overloaded.module
    def torch(module):
        """
        We use the @overloaded.module to specify we're writing here
        a function which should overload the function with the same
        name in the <torch> module
        :param module: object which stores the overloading functions

        Note that we used the @staticmethod decorator as we're in a
        class
        """

        def add(x, y):
            """
            You can write the function to overload in the most natural
            way, so this will be called whenever you call torch.add on
            Logging Tensors, and the x and y you get are also Logging
            Tensors, so compared to the @overloaded.method, you see
            that the @overloaded.module does not hook the arguments.
            """
            print("Log function torch.add")
            return x + y

        # Just register it using the module variable
        module.add = add

        @overloaded.function
        def mul(x, y):
            """
            You can also add the @overloaded.function decorator to also
            hook arguments, ie all the LoggingTensor are replaced with
            their child attribute
            """
            print("Log function torch.mul")
            return x * y

        # Just register it using the module variable
        module.mul = mul

        # You can also overload functions in submodules!
        @overloaded.module
        def nn(module):
            """
            The syntax is the same, so @overloaded.module handles recursion
            Note that we don't need to add the @staticmethod decorator
            """

            @overloaded.module
            def functional(module):
                def relu(x):
                    print("Log function torch.nn.functional.relu")
                    return x * (x.child > 0)

                module.relu = relu

            module.functional = functional

        # Modules should be registered just like functions
        module.nn = nn

    @classmethod
    def on_function_call(cls, command):
        """
        Override this to perform a specific action for each call of a torch
        function with arguments containing syft tensors of the class doing
        the overloading
        """
        cmd, _, args, kwargs = command
        print("Default log", cmd)

    @staticmethod
    def simplify(tensor: "LoggingTensor") -> tuple:
        """
        This function takes the attributes of a LogTensor and saves them in a tuple
        Args:
            tensor (LoggingTensor): a LogTensor
        Returns:
            tuple: a tuple holding the unique attributes of the log tensor
        Examples:
            data = _simplify(tensor)
        """

        chain = None
        if hasattr(tensor, "child"):
            chain = sy.serde._simplify(tensor.child)
        return (tensor.id, chain)

    @staticmethod
    def detail(worker: AbstractWorker, tensor_tuple: tuple) -> "LoggingTensor":
        """
        This function reconstructs a LogTensor given it's attributes in form of a tuple.
        Args:
            worker: the worker doing the deserialization
            tensor_tuple: a tuple holding the attributes of the LogTensor
        Returns:
            LoggingTensor: a LogTensor
        Examples:
            logtensor = detail(data)
        """
        obj_id, chain = tensor_tuple

        tensor = LoggingTensor(owner=worker, id=obj_id)

        if chain is not None:
            chain = sy.serde._detail(worker, chain)
            tensor.child = chain

        return tensor



from typing import Callable
from typing import Dict
from typing import List
from typing import Tuple

import numpy as np

from syft.frameworks.torch.tensors.decorators.logging import LoggingTensor
from syft.frameworks.torch.tensors.interpreters.autograd import AutogradTensor
from syft.generic.frameworks.types import FrameworkTensorType
from syft.workers.abstract import AbstractWorker

from syft import exceptions

hook_method_args_functions = {}
hook_method_response_functions = {}
get_tensor_type_functions = {}

base_types = {int, float, str, bool, bytes, bytearray, complex}

one = lambda _args: 1
get_child = lambda i: i.child

### Hook Args Registries ###
# If you have a type that will be fed into hooked functions, you must add it to
# these registeries using the public functions in the Registration Logic
# section below.
# WARNING: Do not attempt to manipulate them by hand. These registries should
#    not be used outside of this module. Use the helper functions instead.

# dict to specify the action depending of the type found
type_rule = {
    list: lambda _args: [build_rule(a) for a in _args],
    tuple: lambda _args: tuple([build_rule(a) for a in _args]),
    dict: one,  # FIXME This is for additiveShareTensor.child, it can be confusing and AST.child
    np.ndarray: one,
    # should perhaps be of type ShareDict extending dict or something like this
    LoggingTensor: one,
    AutogradTensor: one,
}

# Dict to return the proper lambda function for the right framework or syft tensor type
forward_func = {
    LoggingTensor: get_child,
    AutogradTensor: get_child,
    "my_syft_tensor_type": get_child,
}

# Dict to return the proper lambda function for the right framework or syft tensor type
backward_func = {
    LoggingTensor: lambda i: LoggingTensor().on(i, wrap=False),
    AutogradTensor: lambda i: AutogradTensor(data=i).on(i, wrap=False),
    "my_syft_tensor_type": lambda i, **kwargs: "my_syft_tensor_type(**kwargs).on(i, wrap=False)",
}

# Methods or functions whose signature changes a lot and that we don't want to "cache", because
# they have an arbitrary number of tensors in args which can trigger unexpected behaviour
ambiguous_methods = set()
ambiguous_functions = set()


### Registration logic ###
def register_type_rule(new_type_rules: Dict):
    global type_rule
    type_rule = {**type_rule, **new_type_rules}


def register_forward_func(new_forward_rules: Dict):
    global forward_func
    forward_func = {**forward_func, **new_forward_rules}


def register_backward_func(new_backward_rules: Dict):
    global backward_func
    backward_func = {**backward_func, **new_backward_rules}


def register_ambiguous_method(*method):
    global ambiguous_methods
    ambiguous_methods.update(set(method))


def register_ambiguous_function(*function):
    global ambiguous_functions
    ambiguous_functions.update(set(function))


def default_backward_func(tensorcls):
    return lambda i, **kwargs: tensorcls(**kwargs).on(i, wrap=False)


def default_register_tensor(*tensorcls):
    register_type_rule({t: one for t in tensorcls})
    register_forward_func({t: get_child for t in tensorcls})
    register_backward_func({t: default_backward_func(t) for t in tensorcls})


### Main hook args implementation ###


def unwrap_args_from_method(attr, method_self, args, kwargs):
    """Method arguments are sometimes simple types (such as strings or ints) but sometimes
    they are custom Syft tensors such as wrappers (i.e. FrameworkTensor), LoggingTensor
    or some other tensor type. Complex types (which have a .child attribute) need to
    have arguments converted from the arg to arg.child so that the types match as the
    method is being called down the chain. To make this efficient, we cache which args
    need to be replaced with their children in a dictionary called
    hook_method_args_functions. However, sometimes a method (an attr) has multiple
    different argument signatures, such that sometimes arguments have .child objects
    and other times they don't (such as x.div(), which can accept either a tensor or a
    float as an argument). This invalidates the cache, so we need to have a try/except
    which refreshes the cache if the signature triggers an error.

    Args:
        attr (str): the name of the method being called
        method_self: the tensor on which the method is being called
        args (list): the arguments being passed to the method
        kwargs (dict): the keyword arguments being passed to the function
            (these are not hooked ie replace with their .child attr)
    """
    # Specify an id to distinguish methods from different classes
    # As they won't be used with the same arg types
    attr_id = type(method_self).__name__ + "." + attr

    try:
        assert attr not in ambiguous_methods

        # Load the utility function to transform the args
        hook_args = hook_method_args_functions[attr_id]
        # Try running it
        new_self, new_args = hook_args((method_self, args))

    except (IndexError, KeyError, AssertionError):  # Update the function in case of an error
        args_hook_function, _ = build_unwrap_args_from_function((method_self, args))
        # Store this utility function in the registry
        hook_method_args_functions[attr_id] = args_hook_function
        # Run it
        new_self, new_args = args_hook_function((method_self, args))

    return new_self, new_args, kwargs


def unwrap_args_from_function(attr, args, kwargs, return_args_type=False):
    """See unwrap_args_from_method for details

    Args:
        attr (str): the name of the function being called
        args (list): the arguments being passed to the function
        kwargs (dict): the keyword arguments being passed to the function
            (these are not hooked ie replace with their .child attr)
        return_args_type (bool): return the type of the tensors in the
        original arguments

    Returns:
        - the arguments where all tensors are replaced with their child
        - the type of this new child
        (- the type of the tensors in the arguments)
    """
    try:
        assert attr not in ambiguous_functions
        # Load the utility function to transform the args
        # TODO rename registry or use another one than for methods
        hook_args = hook_method_args_functions[attr]
        get_tensor_type_function = get_tensor_type_functions[attr]
        # Try running it
        new_args = hook_args(args)

    except (IndexError, KeyError, AssertionError):  # Update the function in case of an error
        args_hook_function, get_tensor_type_function = build_unwrap_args_from_function(
            args, return_tuple=True
        )
        # Store the utility functions in registries
        hook_method_args_functions[attr] = args_hook_function
        get_tensor_type_functions[attr] = get_tensor_type_function
        # Run it
        new_args = args_hook_function(args)

    new_type = get_tensor_type_function(new_args)
    if return_args_type:
        args_type = get_tensor_type_function(args)
        return new_args, kwargs, new_type, args_type
    else:
        return new_args, kwargs, new_type


def build_unwrap_args_from_function(args, return_tuple=False):
    """
    Build the function f that hook the arguments:
    f(args) = new_args
    """
    # Inspect the call to find tensor arguments and return a rule whose
    # structure is the same as the args object, with 1 where there was
    # (framework or syft) tensors and 0 when not (ex: number, str, ...)
    rule = build_rule(args)
    # Build a function with this rule to efficiently replace syft tensors
    # (but not pointer) with their child in the args objects
    args_hook_function = build_unwrap_args_with_rules(args, rule, return_tuple)
    # Build a function with this rule to efficiently the child type of the
    # tensor found in the args
    get_tensor_type_function = build_get_tensor_type(rule)
    return args_hook_function, get_tensor_type_function


def hook_response(attr, response, wrap_type, wrap_args={}, new_self=None):
    """
    When executing a command, arguments are inspected and all tensors are replaced
    with their child attribute until a pointer or a framework tensor is found (for
    example an argument could be a framework wrapper with a child being a LoggingTensor, with
    a child being a framework tensor). When the result of the command is calculated,
    we need to rebuild this chain in the reverse order (in our example put back
    a LoggingTensor on top of the result and then a framework wrapper).
    To make this efficient, we cache which elements of the response (which can be more
    complicated with nested tuples for example) need to be wrapped in a dictionary called
    hook_method_response_functions. However, sometimes a method (an attr) has multiple
    different response signatures. This invalidates the cache, so we need to have a
    try/except which refreshes the cache if the signature triggers an error.

    Args:
        attr (str): the name of the method being called
        response (list or dict): the arguments being passed to the tensor
        wrap_type (type): the type of wrapper we'd like to have
        wrap_args (dict): options to give to the wrapper (for example the
        precision for the precision tensor)
        new_self: used for the can just below of inplace ops
    """

    # inline methods should just return new_self
    if "__i" == attr[0:3]:
        return new_self

    # TODO: Why do we need to cast it in a tuple? this is a (small) time waste
    response_is_tuple = isinstance(response, tuple)

    # Add an artificial tuple
    if not response_is_tuple:
        response = (response, 1)

    hash_wrap_args = hash(frozenset(wrap_args.items()))
    attr_id = f"{attr}@{wrap_type.__name__}.{response_is_tuple}.{hash_wrap_args}"

    try:
        assert attr not in ambiguous_functions

        # Load the utility function to transform the args
        response_hook_function = hook_method_response_functions[attr_id]
        # Try running it
        new_response = response_hook_function(response)

    except (IndexError, KeyError, AssertionError):  # Update the function in case of an error
        response_hook_function = build_wrap_reponse_from_function(response, wrap_type, wrap_args)
        # Store this utility function in the registry
        hook_method_response_functions[attr_id] = response_hook_function
        # Run it
        new_response = response_hook_function(response)

    # Remove the artificial tuple
    if not response_is_tuple:
        new_response, _ = new_response

    return new_response


def build_wrap_reponse_from_function(response, wrap_type, wrap_args):
    """
    Build the function that hook the response.

    Example:
        p is of type Pointer
        f is the hook_response_function
        then f(p) = (Wrapper)>Pointer
    """
    # Inspect the call to find tensor arguments and return a rule whose
    # structure is the same as the response object, with 1 where there was
    # (framework or syft) tensors and 0 when not (ex: number, str, ...)
    rule = build_rule(response)
    # Build a function with this rule to efficiently replace syft tensors
    # (but not pointer) with their child in the args objects
    response_hook_function = build_wrap_response_with_rules(response, rule, wrap_type, wrap_args)
    return response_hook_function


def build_rule(args):
    """
    Inspect the args object to find framework or syft tensor arguments and
    return a rule whose structure is the same as the args object,
    with 1 where there was (framework or syft) tensors and 0 when
    not (ex: number, str, ...)

    Example:
        in: ([tensor(1, 2), Pointer@bob], 42)
        out: ([1, 1], 0)
    """

    type_args = type(args)
    # for list, tuple but also tensors and syft tensors
    if type_args in type_rule:
        return type_rule[type_args](args)
    # for int, float, str, etc
    elif type_args in base_types:
        return 0
    else:
        # New kind of return with pytorch 1.1
        if "torch.return_types" in str(type_args):
            return type_rule[tuple](args)
        # Still remain ellipsis, slices, etc.
        return 0


def build_unwrap_args_with_rules(args, rules, return_tuple=False):
    """
    Build a function given some rules to efficiently replace in the args object
    syft tensors with their child (but not pointer as they don't have .child),
    and do nothing for other type of object including framework tensors, str,
    numbers, bool, etc.
    Pointers trigger an error which can be caught to get the location for
    forwarding the call.

    Args:
        args (tuple): the arguments given to the function / method
        rules (tuple): the same structure but with boolean, true when there is
            a tensor
        return_tuple (bool): force to return a tuple even with a single element

    Return:
        a function that replace syft arg in args with arg.child
    """

    # get the transformation lambda for each args
    lambdas = [
        typed_identity(a)  # return the same obj with an identity fct with a type check if needed
        if not r  # if the rule is a number == 0.
        else build_unwrap_args_with_rules(
            a, r, True
        )  # If not, call recursively build_unwrap_args_with_rules
        if isinstance(r, (list, tuple))  # if the rule is a list or tuple.
        # Last if not, rule is probably == 1 so use type to return the right transformation.
        else lambda i: forward_func[type(i)](i)
        for a, r in zip(args, rules)  # And do this for all the args / rules provided
    ]

    # Instead of iterating which is slow, we use trick to efficiently
    # apply each lambda to each arg
    folds = {
        0: zero_fold,
        1: one_fold(return_tuple),
        2: two_fold,
        3: three_fold,
        4: four_fold,
        5: five_fold,
        6: six_fold,
        7: seven_fold,
        8: eight_fold,
    }
    try:
        f = folds[len(lambdas)]
    except KeyError:
        f = many_fold

    return lambda x: f(lambdas, x)


def build_get_tensor_type(rules, layer=None):
    """
    Build a function which uses some rules to find efficiently the first tensor in
    the args objects and return the type of its child.

    Args:
        rules (tuple): a skeleton object with the same structure as args but each tensor
            is replaced with a 1 and other types (int, str) with a 0
        layer (list or None): keep track of the path of inspection: each element in the list
            stand for one layer of deepness into the object, and its value for the index
            in the current layer. See example for details

    Returns:
        a function returning a type

    Example:
        *Understanding the layer parameter*
        obj = (a, [b, (c, d)], e)
        the layer position is for:
        a: [0]
        b: [1, 0]
        c: [1, 1, 0]
        d: [1, 1, 1]
        e: [2]

        *Global behaviour example*
        rules = (0, [1, (0, 0), 0)
        - First recursion level
          0 found -> do nothing
          list found -> recursive call with layer = [1]
        - Second recursion level
          1 found -> update layer to [1, 0]
                     build the function x: type(x[1][0])
                     break
        - Back to first recursion level
          save the function returned in the lambdas list
          0 found -> do nothing
          exit loop
          return the first (and here unique) function


    """
    # We keep note of the first layer or recursion level to return at the end
    # only one function and instantiate the layer list the first time
    first_layer = layer is None

    if first_layer:
        layer = []

    # Iteration through the rules object
    lambdas = []
    for i, r in enumerate(rules):
        if r == 1:  # if a tensor is found
            layer.append(i)
            lambdas.append(
                # the layer object is given to build a getter to reach the
                # tensor position and then the type() is called on the obj found
                lambda a: type(get_element_at[len(layer)](*layer)(a))
            )
            # we only need one to get the type of all tensors as they should be the same
            break
        if isinstance(r, (list, tuple)):  # we iterate recursively if necessary
            layer.append(i)
            lambdas += build_get_tensor_type(r, layer)

    if first_layer:
        try:
            return lambdas[0]
        except IndexError:
            # Some functions don't have tensors in their signature so rules is only made of 0s,
            # Hence lambdas is empty. Raising PureFrameworkTensorFoundError triggers an execution of
            # the un-hooked (so native) function which is perfect in that case.
            raise exceptions.PureFrameworkTensorFoundError
    else:
        return lambdas


# Function helpers to convert [a, b, c, ...] -> obj[a][b][c][...]
def one_layer(idx1):
    return lambda l: l[idx1]


def two_layers(idx1, idx2):
    return lambda l: one_layer(idx2)(l[idx1])


def three_layers(idx1, *ids):
    return lambda l: two_layers(*ids)(l[idx1])


def four_layers(idx1, *ids):
    return lambda l: three_layers(*ids)(l[idx1])


get_element_at = {1: one_layer, 2: two_layers, 3: three_layers, 4: four_layers}


def build_wrap_response_with_rules(response, rules, wrap_type, wrap_args, return_tuple=False):
    """
    Build a function given some rules to efficiently replace in the response object
    syft or framework tensors with a wrapper, and do nothing for other types of object
    including , str, numbers, bool, etc.

    Args:
        response: a response used to build the hook function
        rules: the same structure objects but with boolean, at true when is replaces
            a tensor
        return_tuple: force to return a tuple even with a single element

    Response:
        a function to "wrap" the response
    """

    # get the transformation lambda for each args
    lambdas = [
        (lambda i: i)  # return the same object
        if not r  # if the rule is a number == 0.
        else build_wrap_response_with_rules(
            a, r, wrap_type, wrap_args, True
        )  # If not, call recursively build_wrap_response_with_rules
        if isinstance(r, (list, tuple))  # if the rule is a list or tuple.
        # Last if not, rule is probably == 1 so use type to return the right transformation.
        else lambda i: backward_func[wrap_type](i, **wrap_args)
        for a, r in zip(response, rules)  # And do this for all the responses / rules provided
    ]

    # Instead of iterating which is slow, we use trick to efficiently
    # apply each lambda to each arg
    folds = {
        0: zero_fold,
        1: one_fold(return_tuple),
        2: two_fold,
        3: three_fold,
        4: four_fold,
        5: five_fold,
        6: six_fold,
        7: seven_fold,
        8: eight_fold,
    }
    try:
        f = folds[len(lambdas)]
    except KeyError:
        f = many_fold

    return lambda x: f(lambdas, x)


def zero_fold(*a, **k):
    return tuple()


def one_fold(return_tuple, **kwargs):
    def _one_fold(lambdas, args, **kwargs):
        return lambdas[0](args[0], **kwargs)

    def tuple_one_fold(lambdas, args):
        return (lambdas[0](args[0], **kwargs),)

    return {False: _one_fold, True: tuple_one_fold}[return_tuple]


def two_fold(lambdas, args, **kwargs):
    return lambdas[0](args[0], **kwargs), lambdas[1](args[1], **kwargs)


def three_fold(lambdas, args, **kwargs):
    return (
        lambdas[0](args[0], **kwargs),
        lambdas[1](args[1], **kwargs),
        lambdas[2](args[2], **kwargs),
    )


def four_fold(lambdas, args, **kwargs):
    return (
        lambdas[0](args[0], **kwargs),
        lambdas[1](args[1], **kwargs),
        lambdas[2](args[2], **kwargs),
        lambdas[3](args[3], **kwargs),
    )


def five_fold(lambdas, args, **kwargs):
    return (
        lambdas[0](args[0], **kwargs),
        lambdas[1](args[1], **kwargs),
        lambdas[2](args[2], **kwargs),
        lambdas[3](args[3], **kwargs),
        lambdas[4](args[4], **kwargs),
    )


def six_fold(lambdas, args, **kwargs):
    return (
        lambdas[0](args[0], **kwargs),
        lambdas[1](args[1], **kwargs),
        lambdas[2](args[2], **kwargs),
        lambdas[3](args[3], **kwargs),
        lambdas[4](args[4], **kwargs),
        lambdas[5](args[5], **kwargs),
    )


def seven_fold(lambdas, args, **kwargs):
    return (
        lambdas[0](args[0], **kwargs),
        lambdas[1](args[1], **kwargs),
        lambdas[2](args[2], **kwargs),
        lambdas[3](args[3], **kwargs),
        lambdas[4](args[4], **kwargs),
        lambdas[5](args[5], **kwargs),
        lambdas[6](args[6], **kwargs),
    )


def eight_fold(lambdas, args, **kwargs):
    return (
        lambdas[0](args[0], **kwargs),
        lambdas[1](args[1], **kwargs),
        lambdas[2](args[2], **kwargs),
        lambdas[3](args[3], **kwargs),
        lambdas[4](args[4], **kwargs),
        lambdas[5](args[5], **kwargs),
        lambdas[6](args[6], **kwargs),
        lambdas[7](args[7], **kwargs),
    )


def many_fold(lambdas, args, **kwargs):
    return tuple([lambdas[i](args[i], **kwargs) for i in range(len(lambdas))])


# Add the possibility to make a type check in the identity function applied
# On some arg which could be None are of another type.
# Could add more checks but not sure it is needed so far.


def typed_identity(a):
    """
    We need to add typed identity for arguments which can be either number
    or tensors. If the argument changes from an int to a tensor, the
    assertion error triggered by typed_identity will be caught and a
    new signature will be computed for the command.
    """
    if a is None:

        def none_identity(i):
            assert i is None
            return i

        return none_identity

    elif type(a) in (int, float, bool):

        def number_identity(i):
            assert isinstance(i, type(a))
            return i

        return number_identity

    else:
        return lambda i: i


# -- Fast way to register responses and transform tensors in pointers

register_response_functions = {}


def register_response(
    attr: str, response: object, response_ids: object, owner: AbstractWorker
) -> object:
    """
    When a remote worker execute a command sent by someone else, the response is
    inspected: all tensors are stored by this worker and a Pointer tensor is
    made for each of them.

    To make this efficient, we cache which elements of the response (which can be more
    complicated with nested tuples for example) in the dict register_response_functions

    However, sometimes a function  (an attr) has multiple different response signatures.
    This invalidates the cache, so we need to have a try/except which refreshes the
    cache if the signature triggers an error.

    Args:
        attr (str): the name of the function being called
        response (object): the response of this function
        owner (BaseWorker): the worker which registers the tensors
    """

    # TODO: Why do we need to cast it in a tuple? this is a (small) time waste
    response_is_tuple = isinstance(response, tuple)

    # Add an artificial tuple
    if not response_is_tuple:
        response = (response, 1)

    attr_id = "{}".format(attr)

    try:
        assert attr not in ambiguous_functions

        # Load the utility function to register the response and transform tensors with pointers
        register_response_function = register_response_functions[attr_id]
        # Try running it
        new_response = register_response_function(response, response_ids=response_ids, owner=owner)

    except (IndexError, KeyError, AssertionError):  # Update the function in cas of an error
        register_response_function = build_register_response_function(response)
        # Store this utility function in the registry
        register_response_functions[attr_id] = register_response_function
        # Run it
        new_response = register_response_function(response, response_ids=response_ids, owner=owner)

    # Remove the artificial tuple
    if not response_is_tuple:
        new_response, _ = new_response

    return new_response


def build_register_response_function(response: object) -> Callable:
    """
    Build the function that registers the response and replaces tensors with pointers.

    Example:
        (1, tensor([1, 2]) is the response
        f is the register_response_function
        then f(p) = (1, (Wrapper)>Pointer)
    """
    # Inspect the call to find tensor arguments and return a rule whose
    # structure is the same as the response object, with 1 where there was
    # (framework or syft) tensors and 0 when not (ex: number, str, ...)
    rule = build_rule(response)
    # Build a function with this rule to efficiently replace syft tensors
    # (but not pointer) with their child in the args objects
    response_hook_function = build_register_response(response, rule)
    return response_hook_function


def register_tensor(
    tensor: FrameworkTensorType, owner: AbstractWorker, response_ids: List = list()
):
    """
    Registers a tensor.

    Args:
        tensor: A tensor.
        owner: The owner that makes the registration.
        response_ids: List of ids where the tensor should be stored
            and each id is pop out when needed.
    """
    tensor.owner = owner
    try:
        tensor.id = response_ids.pop(-1)
    except IndexError:
        raise exceptions.ResponseSignatureError
    owner.register_obj(tensor)


def build_register_response(response: object, rules: Tuple, return_tuple: bool = False) -> Callable:
    """
    Build a function given some rules to efficiently replace in the response object
    framework tensors with a pointer after they are registered, and do nothing for other
    types of object including , str, numbers, bool, etc.

    Args:
        response: the response
        rules: the rule specifying where the tensors are
        return_tuple: force to return a tuple even with a single element
    Returns:
        The function to apply on generic responses
    """

    # get the transformation lambda for each args
    lambdas = [
        (lambda i, **kwargs: i)  # return the same object
        if not r or not hasattr(a, "owner")  # if the rule is a number == 0.
        else build_register_response(
            a, r, True
        )  # If not, call recursively build_wrap_response_with_rules
        if isinstance(r, (list, tuple))  # if the rule is a list or tuple.
        # Last if not, rule is probably == 1 so use type to return the right transformation.
        else lambda i, **kwargs: register_tensor(i, **kwargs)
        for a, r in zip(response, rules)  # And do this for all the responses / rules provided
    ]

    # Instead of iterating which is slow, we use trick to efficiently
    # apply each lambda to each arg
    folds = {
        0: zero_fold,
        1: one_fold(return_tuple),
        2: two_fold,
        3: three_fold,
        4: four_fold,
        5: five_fold,
        6: six_fold,
        7: seven_fold,
        8: eight_fold,
    }
    try:
        f = folds[len(lambdas)]
    except KeyError:
        f = many_fold

    return lambda x, **kwargs: f(lambdas, x, **kwargs)



import torch
import syft


def test__str__():
    t = torch.tensor([[3, 9], [4, 1]])
    crt = t.fix_precision(field_type="int100", precision_fractional=2, storage="crt")

    assert isinstance(crt.__str__(), str)


def test_eq():
    t_a = torch.tensor([[3, 9], [4, 1]])
    t_b = torch.tensor([[3, 9], [4, 1]])
    t_c = torch.tensor([[2, 9], [4, 2]])

    crt_a = t_a.fix_precision(field_type="int100", precision_fractional=2, storage="crt")
    crt_b = t_b.fix_precision(field_type="int100", precision_fractional=2, storage="crt")
    crt_c = t_c.fix_precision(field_type="int100", precision_fractional=2, storage="crt")

    eq_ab = (crt_a == crt_b).float_precision()
    eq_ac = (crt_a == crt_c).float_precision()

    assert (eq_ab == torch.tensor([[1.0, 1.0], [1.0, 1.0]])).all()
    assert (eq_ac == torch.tensor([[0.0, 1.0], [1.0, 0.0]])).all()


def test__neg__():
    t = torch.tensor([[1, -3], [3, -2]])
    crt = t.fix_precision(field_type="int100", precision_fractional=2, storage="crt")

    neg = -crt

    assert (neg.float_precision() == torch.tensor([[-1.0, 3.0], [-3.0, 2.0]])).all()


def test_add():
    t_a = torch.tensor([[1, 2], [3, -4]])
    t_b = torch.tensor([[1, -3], [3, -2]])

    crt_a = t_a.fix_precision(field_type="int100", precision_fractional=2, storage="crt")
    crt_b = t_b.fix_precision(field_type="int100", precision_fractional=2, storage="crt")

    result = crt_a + crt_b

    assert (result.float_precision() == torch.tensor([[2.0, -1.0], [6.0, -6.0]])).all()

    # With scalars
    result = crt_a + 1

    assert (result.float_precision() == torch.tensor([[2.0, 3.0], [4.0, -3.0]])).all()


def test_sub():
    t_a = torch.tensor([[1, 2], [5, -4]])
    t_b = torch.tensor([[1, -3], [3, -2]])

    crt_a = t_a.fix_precision(field_type="int100", precision_fractional=2, storage="crt")
    crt_b = t_b.fix_precision(field_type="int100", precision_fractional=2, storage="crt")

    result = crt_a - crt_b

    assert (result.float_precision() == torch.tensor([[0.0, 5.0], [2.0, -2.0]])).all()

    # With scalars
    result_a = crt_a - 1
    result_b = 1 - crt_a

    assert (result_a.float_precision() == torch.tensor([[0.0, 1.0], [4.0, -5.0]])).all()
    assert (result_b.float_precision() == torch.tensor([[0.0, -1.0], [-4.0, 5.0]])).all()


def test_mul():
    t_a = torch.tensor([[1, -2], [3, -2]])
    t_b = torch.tensor([[1, 2], [3, -2]])

    crt_a = t_a.fix_precision(field_type="int100", precision_fractional=2, storage="crt")
    crt_b = t_b.fix_precision(field_type="int100", precision_fractional=2, storage="crt")

    result = crt_a * crt_b

    assert (result.float_precision() == torch.tensor([[1.0, -4.0], [9.0, 4.0]])).all()

    # With scalar
    result = 3 * crt_a

    assert (result.float_precision() == torch.tensor([[3.0, -6.0], [9.0, -6.0]])).all()


def test_send_and_get(workers):
    alice, bob = (workers["alice"], workers["bob"])

    t = torch.tensor([1, 2])
    crt = t.fix_precision(field_type="int100", precision_fractional=2, storage="crt")

    to_alice = crt.send(alice)
    to_alice_id = to_alice.id_at_location

    assert to_alice_id in alice._objects

    to_bob_to_alice = to_alice.send(bob)
    to_bob_to_alice_id = to_bob_to_alice.id_at_location

    assert to_alice_id in alice._objects
    assert to_bob_to_alice_id in bob._objects

    to_alice_back = to_bob_to_alice.get()

    assert to_bob_to_alice_id not in bob._objects
    assert to_alice_id in alice._objects

    t_back = to_alice_back.get()

    assert to_alice_id not in alice._objects

    eq = (t_back == crt).float_precision()
    assert (eq == torch.tensor([1.0, 1.0])).all()


def test_share_and_get(workers):
    alice, bob, james = (workers["alice"], workers["bob"], workers["james"])

    t = torch.tensor([1, 2])
    crt = t.fix_precision(field_type="int100", precision_fractional=2, storage="crt")
    copy = t.fix_precision(field_type="int100", precision_fractional=2, storage="crt")

    shared = crt.share(alice, bob, crypto_provider=james)
    back = shared.get()

    assert (back.float_precision() == copy.float_precision()).all()



import pytest
import torch
import syft as sy
from syft.frameworks.torch.linalg import inv_sym


def test_inv_sym(hook, workers):
    """
    Testing inverse of symmetric matrix with MPC
    """
    torch.manual_seed(42)  # Truncation might not always work so we set the random seed

    bob = workers["bob"]
    alice = workers["alice"]
    crypto_prov = sy.VirtualWorker(hook, id="crypto_prov")

    x = torch.Tensor([[0.4627, 0.8224], [0.8224, 2.4084]])

    x = x.fix_precision(precision_fractional=6).share(bob, alice, crypto_provider=crypto_prov)
    gram = x.matmul(x.t())
    gram_inv = inv_sym(gram)

    gram_inv = gram_inv.get().float_precision()
    gram = gram.get().float_precision()

    diff = (gram_inv - gram.inverse()).abs()
    assert (diff < 1e-3).all()



import torch as th
from torch.utils.data import BatchSampler, RandomSampler, SequentialSampler
import numpy as np

from syft.generic.object_storage import ObjectStorage
from syft.federated.train_config import TrainConfig


class FederatedClient(ObjectStorage):
    """A Client able to execute federated learning in local datasets."""

    def __init__(self, datasets=None):
        super().__init__()
        self.datasets = datasets if datasets is not None else dict()
        self.optimizer = None
        self.train_config = None

    def add_dataset(self, dataset, key: str):
        if key not in self.datasets:
            self.datasets[key] = dataset
        else:
            raise ValueError(f"Key {key} already exists in Datasets")

    def remove_dataset(self, key: str):
        if key in self.datasets:
            del self.datasets[key]

    def set_obj(self, obj: object):
        """Registers objects checking if which objects it should cache.

        Args:
            obj: An object to be registered.
        """
        if isinstance(obj, TrainConfig):
            self.train_config = obj
            self.optimizer = None
        else:
            super().set_obj(obj)

    def _check_train_config(self):
        if self.train_config is None:
            raise ValueError("Operation needs TrainConfig object to be set.")

    def _build_optimizer(
        self, optimizer_name: str, model, optimizer_args: dict
    ) -> th.optim.Optimizer:
        """Build an optimizer if needed.

        Args:
            optimizer_name: A string indicating the optimizer name.
            optimizer_args: A dict containing the args used to initialize the optimizer.
        Returns:
            A Torch Optimizer.
        """
        if self.optimizer is not None:
            return self.optimizer

        if optimizer_name in dir(th.optim):
            optimizer = getattr(th.optim, optimizer_name)
            self.optimizer = optimizer(model.parameters(), **optimizer_args)
        else:
            raise ValueError("Unknown optimizer: {}".format(optimizer_name))
        return self.optimizer

    def fit(self, dataset_key: str, **kwargs):
        """Fits a model on the local dataset as specified in the local TrainConfig object.

        Args:
            dataset_key: Identifier of the local dataset that shall be used for training.
            **kwargs: Unused.

        Returns:
            loss: Training loss on the last batch of training data.
        """
        self._check_train_config()

        if dataset_key not in self.datasets:
            raise ValueError("Dataset {} unknown.".format(dataset_key))

        model = self.get_obj(self.train_config._model_id).obj
        loss_fn = self.get_obj(self.train_config._loss_fn_id).obj

        self._build_optimizer(
            self.train_config.optimizer, model, optimizer_args=self.train_config.optimizer_args
        )

        return self._fit(model=model, dataset_key=dataset_key, loss_fn=loss_fn)

    def _create_data_loader(self, dataset_key: str, shuffle: bool = False):
        data_range = range(len(self.datasets[dataset_key]))
        if shuffle:
            sampler = RandomSampler(data_range)
        else:
            sampler = SequentialSampler(data_range)
        data_loader = th.utils.data.DataLoader(
            self.datasets[dataset_key],
            batch_size=self.train_config.batch_size,
            sampler=sampler,
            num_workers=0,
        )
        return data_loader

    def _fit(self, model, dataset_key, loss_fn):
        model.train()
        data_loader = self._create_data_loader(
            dataset_key=dataset_key, shuffle=self.train_config.shuffle
        )

        loss = None
        iteration_count = 0

        for _ in range(self.train_config.epochs):
            for (data, target) in data_loader:
                # Set gradients to zero
                self.optimizer.zero_grad()

                # Update model
                output = model(data)
                loss = loss_fn(target=target, pred=output)
                loss.backward()
                self.optimizer.step()

                # Update and check interation count
                iteration_count += 1
                if iteration_count >= self.train_config.max_nr_batches >= 0:
                    break

        return loss

    def evaluate(
        self,
        dataset_key: str,
        return_histograms: bool = False,
        nr_bins: int = -1,
        return_loss: bool = True,
        return_raw_accuracy: bool = True,
    ):
        """Evaluates a model on the local dataset as specified in the local TrainConfig object.

        Args:
            dataset_key: Identifier of the local dataset that shall be used for training.
            return_histograms: If True, calculate the histograms of predicted classes.
            nr_bins: Used together with calculate_histograms. Provide the number of classes/bins.
            return_loss: If True, loss is calculated additionally.
            return_raw_accuracy: If True, return nr_correct_predictions and nr_predictions

        Returns:
            Dictionary containing depending on the provided flags:
                * loss: avg loss on data set, None if not calculated.
                * nr_correct_predictions: number of correct predictions.
                * nr_predictions: total number of predictions.
                * histogram_predictions: histogram of predictions.
                * histogram_target: histogram of target values in the dataset.
        """
        self._check_train_config()

        if dataset_key not in self.datasets:
            raise ValueError("Dataset {} unknown.".format(dataset_key))

        eval_result = dict()
        model = self.get_obj(self.train_config._model_id).obj
        loss_fn = self.get_obj(self.train_config._loss_fn_id).obj
        model.eval()
        device = "cpu"
        data_loader = self._create_data_loader(dataset_key=dataset_key, shuffle=False)
        test_loss = 0.0
        correct = 0
        if return_histograms:
            hist_target = np.zeros(nr_bins)
            hist_pred = np.zeros(nr_bins)

        with th.no_grad():
            for data, target in data_loader:
                data, target = data.to(device), target.to(device)
                output = model(data)
                if return_loss:
                    test_loss += loss_fn(output, target).item()  # sum up batch loss
                pred = output.argmax(
                    dim=1, keepdim=True
                )  # get the index of the max log-probability
                if return_histograms:
                    hist, _ = np.histogram(target, bins=nr_bins, range=(0, nr_bins))
                    hist_target += hist
                    hist, _ = np.histogram(pred, bins=nr_bins, range=(0, nr_bins))
                    hist_pred += hist
                if return_raw_accuracy:
                    correct += pred.eq(target.view_as(pred)).sum().item()

        if return_loss:
            test_loss /= len(data_loader.dataset)
            eval_result["loss"] = test_loss
        if return_raw_accuracy:
            eval_result["nr_correct_predictions"] = correct
            eval_result["nr_predictions"] = len(data_loader.dataset)
        if return_histograms:
            eval_result["histogram_predictions"] = hist_pred
            eval_result["histogram_target"] = hist_target

        return eval_result



"""
This file exists to provide one common place for all serialisation and simplify_ and _detail
for all tensors (Torch and Numpy).
"""
from collections import OrderedDict
import io
from tempfile import TemporaryFile
from typing import Tuple, List
import warnings

import numpy
import torch

import syft
from syft.generic.pointers.pointer_tensor import PointerTensor
from syft.generic.tensor import initialize_tensor
from syft.workers.abstract import AbstractWorker


def _serialize_tensor(tensor) -> bin:
    """Serialize the tensor using as default Torch serialization strategy
    This function can be overridden to provide different tensor serialization strategies

    Args
        (torch.Tensor): an input tensor to be serialized

    Returns
        A serialized version of the input tensor

    """
    return torch_tensor_serializer(tensor)


def _deserialize_tensor(tensor_bin) -> torch.Tensor:
    """Deserialize the input tensor passed as parameter into a Torch tensor.
    This function can be overridden to provide different deserialization strategies

    Args
        tensor_bin: A binary representation of a tensor

    Returns
        a Torch tensor
    """
    return torch_tensor_deserializer(tensor_bin)


def numpy_tensor_serializer(tensor: torch.Tensor) -> bin:
    """Strategy to serialize a tensor using numpy npy format.
    If tensor requires to calculate gradients, it will be detached.
    """
    if tensor.requires_grad:
        warnings.warn(
            "Torch to Numpy serializer can only be used with tensors that do not require grad. "
            "Detaching tensor to continue"
        )
        tensor = tensor.detach()

    np_tensor = tensor.numpy()
    outfile = TemporaryFile()
    numpy.save(outfile, np_tensor)
    # Simulate close and open by calling seek
    outfile.seek(0)
    return outfile.read()


def numpy_tensor_deserializer(tensor_bin) -> torch.Tensor:
    """"Strategy to deserialize a binary input in npy format into a Torch tensor"""
    input_file = TemporaryFile()
    input_file.write(tensor_bin)
    # read data from file
    input_file.seek(0)
    return torch.from_numpy(numpy.load(input_file))


def torch_tensor_serializer(tensor) -> bin:
    """Strategy to serialize a tensor using Torch saver"""
    binary_stream = io.BytesIO()
    torch.save(tensor, binary_stream)
    return binary_stream.getvalue()


def torch_tensor_deserializer(tensor_bin) -> torch.Tensor:
    """"Strategy to deserialize a binary input using Torch load"""
    bin_tensor_stream = io.BytesIO(tensor_bin)
    return torch.load(bin_tensor_stream)


# Simplify/Detail Torch Tensors


def _simplify_torch_tensor(tensor: torch.Tensor) -> bin:
    """
    This function converts a torch tensor into a serliaized torch tensor
    using pickle. We choose to use this because PyTorch has a custom and
    very fast PyTorch pickler.

    Args:
        tensor (torch.Tensor): an input tensor to be serialized

    Returns:
        tuple: serialized tuple of torch tensor. The first value is the
        id of the tensor and the second is the binary for the PyTorch
        object. The third is the chain of abstractions, and the fourth
        (optinally) is the chain of graident tensors (nested tuple)
    """

    tensor_bin = _serialize_tensor(tensor)

    # note we need to do this explicitly because torch.save does not
    # seem to be including .grad by default

    if tensor.grad is not None:
        if hasattr(tensor, "child"):
            if isinstance(tensor.child, PointerTensor):
                grad_chain = None
            else:
                grad_chain = _simplify_torch_tensor(tensor.grad)
        else:
            grad_chain = _simplify_torch_tensor(tensor.grad)

    else:
        grad_chain = None

    chain = None

    # I think the pointer bug is is between here

    if hasattr(tensor, "child"):
        chain = syft.serde._simplify(tensor.child)

    # and here... leaving a reerence here so i can find it later
    # TODO fix pointer bug

    tags = tensor.tags
    if tags is not None:
        tags = list(tags)
    return (tensor.id, tensor_bin, chain, grad_chain, tags, tensor.description)


def _detail_torch_tensor(worker: AbstractWorker, tensor_tuple: tuple) -> torch.Tensor:
    """
    This function converts a serialized torch tensor into a torch tensor
    using pickle.

    Args:
        tensor_tuple (bin): serialized obj of torch tensor. It's a tuple where
            the first value is the ID, the second vlaue is the binary for the
            PyTorch object, the third value is the chain of tensor abstractions,
            and the fourth object is the chain of gradients (.grad.grad, etc.)

    Returns:
        torch.Tensor: a torch tensor that was serialized
    """

    tensor_id, tensor_bin, chain, grad_chain, tags, description = tensor_tuple

    tensor = _deserialize_tensor(tensor_bin)

    # note we need to do this explicitly because torch.load does not
    # include .grad informatino
    if grad_chain is not None:
        tensor.grad = _detail_torch_tensor(worker, grad_chain)

    initialize_tensor(
        hook_self=syft.torch.hook,
        cls=tensor,
        is_tensor=True,
        owner=worker,
        id=tensor_id,
        init_args=[],
        kwargs={},
    )

    if tags is not None:

        tags = list(tags)

        for i in range(len(tags)):
            tag = tags[i]
            if isinstance(tag, bytes):
                tag = tag.decode("utf-8")
            tags[i] = tag
        tensor.tags = tags

    if description is not None:
        if isinstance(description, bytes):
            description = description.decode("utf-8")
        tensor.description = description

    if chain is not None:
        chain = syft.serde._detail(worker, chain)
        tensor.child = chain
        tensor.is_wrapper = True

    return tensor


# Simplify/Detail Parameters


def _simplify_torch_parameter(param: torch.nn.Parameter) -> bin:
    """
    This function converts a torch Parameter into a serialized torch Parameter

    Args:
        param (torch.nn.Parameter): an input Parameter to be serialized

    Returns:
        tuple: serialized tuple of torch Parameter. The first value is the
        id of the Parameter and the second is the binary for the PyTorch
        tensor data attribute and last is the requires_grad attr.
    """

    tensor = param.data
    tensor_ser = syft.serde._simplify(tensor)

    grad = param.grad

    if grad is not None and not (hasattr(grad, "child") and isinstance(grad.child, PointerTensor)):
        grad_ser = _simplify_torch_tensor(grad)
    else:
        grad_ser = None

    return (param.id, tensor_ser, param.requires_grad, grad_ser)


def _detail_torch_parameter(worker: AbstractWorker, param_tuple: tuple) -> torch.nn.Parameter:
    """
    This function converts a serialized torch Parameter into a torch Parameter.

    Args:
        param_tuple (tuple): serialized obj of torch tensor. It's a tuple where
            the first value is the ID and the second value is the binary for the
            PyTorch data attribute et and third value is the requires_grad attr.

    Returns:
        torch.Parameter: a torch Parameter that was serialized
    """
    param_id, tensor_ser, requires_grad, grad_ser = param_tuple

    tensor = syft.serde._detail(worker, tensor_ser)

    if grad_ser is not None:
        grad = _detail_torch_tensor(worker, grad_ser)
        grad.garbage_collect_data = False
    elif hasattr(tensor, "child") and isinstance(tensor.child, PointerTensor):
        grad = tensor.attr("grad")
    else:
        grad = None

    param = torch.nn.Parameter(tensor, requires_grad)
    param.id = param_id
    param.grad = grad

    return param


#   Numpy array


def _simplify_ndarray(my_array: numpy.ndarray) -> Tuple[bin, Tuple, str]:
    """
    This function gets the byte representation of the array
        and stores the dtype and shape for reconstruction

    Args:
        my_array (numpy.ndarray): a numpy array

    Returns:
        list: a list holding the byte representation, shape and dtype of the array

    Examples:

        arr_representation = _simplify_ndarray(numpy.random.random([1000, 1000])))

    """
    arr_bytes = my_array.tobytes()
    arr_shape = my_array.shape
    arr_dtype = my_array.dtype.name

    return (arr_bytes, arr_shape, arr_dtype)


def _detail_ndarray(
    worker: AbstractWorker, arr_representation: Tuple[bin, Tuple, str]
) -> numpy.ndarray:
    """
    This function reconstruct a numpy array from it's byte data, the shape and the dtype
        by first loading the byte data with the appropiate dtype and then reshaping it into the
        original shape

    Args:
        worker: the worker doing the deserialization
        arr_representation (tuple): a tuple holding the byte representation, shape
        and dtype of the array

    Returns:
        numpy.ndarray: a numpy array

    Examples:
        arr = _detail_ndarray(arr_representation)

    """
    res = numpy.frombuffer(arr_representation[0], dtype=arr_representation[2]).reshape(
        arr_representation[1]
    )

    assert type(res) == numpy.ndarray

    return res


def _simplify_torch_device(device: torch.device) -> Tuple[str]:
    return device.type


def _detail_torch_device(worker: AbstractWorker, device_type: str) -> torch.device:
    return torch.device(type=device_type)


def _simplify_torch_size(shape: torch.Size) -> Tuple:
    return (list(shape),)


def _detail_torch_size(worker: AbstractWorker, shape: List[int]) -> torch.Size:
    return torch.Size(*shape)


def _simplify_script_module(obj: torch.jit.ScriptModule) -> str:
    """Strategy to serialize a script module using Torch.jit"""
    return obj.save_to_buffer()


def _detail_script_module(worker: AbstractWorker, script_module_bin: str) -> torch.jit.ScriptModule:
    """"Strategy to deserialize a binary input using Torch load"""
    script_module_stream = io.BytesIO(script_module_bin)
    loaded_module = torch.jit.load(script_module_stream)
    return loaded_module


def _simplify_torch_size(size: torch.Size) -> Tuple[int]:
    return tuple(size)


def _detail_torch_size(worker: AbstractWorker, size: Tuple[int]) -> torch.Size:
    return torch.Size(size)


# Maps a type to a tuple containing its simplifier and detailer function
# IMPORTANT: keep this structure sorted A-Z (by type name)
MAP_TORCH_SIMPLIFIERS_AND_DETAILERS = OrderedDict(
    {
        numpy.ndarray: (_simplify_ndarray, _detail_ndarray),
        torch.device: (_simplify_torch_device, _detail_torch_device),
        torch.Size: (_simplify_torch_size, _detail_torch_size),
        torch.jit.ScriptModule: (_simplify_script_module, _detail_script_module),
        torch.jit.TopLevelTracedModule: (_simplify_script_module, _detail_script_module),
        torch.nn.Parameter: (_simplify_torch_parameter, _detail_torch_parameter),
        torch.Tensor: (_simplify_torch_tensor, _detail_torch_tensor),
        torch.Size: (_simplify_torch_size, _detail_torch_size),
    }
)



import torch
import torch as th
import syft

from syft.frameworks.torch.tensors.interpreters.precision import FixedPrecisionTensor
from syft.generic.pointers.pointer_tensor import PointerTensor
import pytest


def test_init(workers):
    pointer = PointerTensor(id=1000, location=workers["alice"], owner=workers["me"])
    pointer.__str__()


def test_create_pointer(workers):
    x = torch.Tensor([1, 2])
    x.create_pointer()


def test_send_default_garbage_collector_true(workers):
    """
    Remote tensor should be garbage collected by default on
    deletion of the Pointer tensor pointing to remote tensor
    """
    x = torch.Tensor([-1, 2])
    x_ptr = x.send(workers["alice"])
    assert x_ptr.child.garbage_collect_data


def test_send_garbage_collect_data_false(workers):
    """
    Remote tensor should be not garbage collected on
    deletion of the Pointer tensor pointing to remote tensor
    """
    x = torch.Tensor([-1, 2])
    x_ptr = x.send(workers["alice"])
    x_ptr.garbage_collection = False
    assert x_ptr.child.garbage_collect_data == False


def test_send_gc_false(workers):
    """
    Remote tensor should be not garbage collected on
    deletion of the Pointer tensor pointing to remote tensor
     """
    x = torch.Tensor([-1, 2])
    x_ptr = x.send(workers["alice"])
    x_ptr.gc = False
    assert x_ptr.child.garbage_collect_data == False
    assert x_ptr.gc == False, "property GC is not in sync"
    assert x_ptr.garbage_collection == False, "property garbage_collection is not in sync"


def test_send_gc_true(workers):
    """
    Remote tensor by default is garbage collected on
    deletion of Pointer Tensor
    """
    x = torch.Tensor([-1, 2])
    x_ptr = x.send(workers["alice"])

    assert x_ptr.gc == True


def test_send_disable_gc(workers):
    """Pointer tensor should be not garbage collected."""
    x = torch.Tensor([-1, 2])
    x_ptr = x.send(workers["alice"]).disable_gc
    assert x_ptr.child.garbage_collect_data == False
    assert x_ptr.gc == False, "property GC is not in sync"
    assert x_ptr.garbage_collection == False, "property garbage_collection is not in sync"


def test_send_get(workers):
    """Test several send get usages"""
    bob = workers["bob"]
    alice = workers["alice"]

    # simple send
    x = torch.Tensor([1, 2])
    x_ptr = x.send(bob)
    x_back = x_ptr.get()
    assert (x == x_back).all()

    # send with variable overwriting
    x = torch.Tensor([1, 2])
    x = x.send(bob)
    x_back = x.get()
    assert (torch.Tensor([1, 2]) == x_back).all()

    # double send
    x = torch.Tensor([1, 2])
    x_ptr = x.send(bob)
    x_ptr_ptr = x_ptr.send(alice)
    x_ptr_back = x_ptr_ptr.get()
    x_back_back = x_ptr_back.get()
    assert (x == x_back_back).all()

    # double send with variable overwriting
    x = torch.Tensor([1, 2])
    x = x.send(bob)
    x = x.send(alice)
    x = x.get()
    x_back = x.get()
    assert (torch.Tensor([1, 2]) == x_back).all()

    # chained double send
    x = torch.Tensor([1, 2])
    x = x.send(bob).send(alice)
    x_back = x.get().get()
    assert (torch.Tensor([1, 2]) == x_back).all()


def test_inplace_send_get(workers):
    tensor = torch.tensor([1.0, -1.0, 3.0, 4.0])
    tensor_ptr = tensor.send_(workers["bob"])

    assert tensor_ptr.id == tensor.id
    assert id(tensor_ptr) == id(tensor)

    tensor_back = tensor_ptr.get_()

    assert tensor_back.id == tensor_ptr.id
    assert tensor_back.id == tensor.id
    assert id(tensor_back) == id(tensor)
    assert id(tensor_back) == id(tensor)

    assert (tensor_back == tensor).all()


def test_repeated_send(workers):
    """Tests that repeated calls to .send(bob) works gracefully.
    Previously garbage collection deleted the remote object
    when .send() was called twice. This test ensures the fix still
    works."""

    # create tensor
    x = torch.Tensor([1, 2])
    print(x.id)

    # send tensor to bob
    x_ptr = x.send(workers["bob"])

    # send tensor again
    x_ptr = x.send(workers["bob"])

    # ensure bob has tensor
    assert x.id in workers["bob"]._objects


def test_remote_autograd(workers):
    """Tests the ability to backpropagate gradients on a remote
    worker."""

    # TEST: simple remote grad calculation

    # create a tensor
    x = torch.tensor([1, 2, 3, 4.0], requires_grad=True)

    # send tensor to bob
    x = x.send(workers["bob"])

    # do some calculation
    y = (x + x).sum()

    # backpropagate on remote machine
    y.backward()

    # check that remote gradient is correct
    x_grad = workers["bob"]._objects[x.id_at_location].grad
    x_grad_target = torch.ones(4).float() + 1
    assert (x_grad == x_grad_target).all()

    # TEST: Ensure remote grad calculation gets properly serded

    # create tensor
    x = torch.tensor([1, 2, 3, 4.0], requires_grad=True).send(workers["bob"])

    # compute function
    y = x.sum()

    # backpropagate
    y.backward()

    # get the gradient created from backpropagation manually
    x_grad = workers["bob"]._objects[x.id_at_location].grad

    # get the entire x tensor (should bring the grad too)
    x = x.get()

    # make sure that the grads match
    assert (x.grad == x_grad).all()


def test_gradient_send_recv(workers):
    """Tests that gradients are properly sent and received along
    with their tensors."""

    # create a tensor
    x = torch.tensor([1, 2, 3, 4.0], requires_grad=True)

    # create gradient on tensor
    x.sum().backward(torch.ones(1))

    # save gradient
    orig_grad = x.grad

    # send and get back
    t = x.send(workers["bob"]).get()

    # check that gradient was properly serde
    assert (t.grad == orig_grad).all()


def test_method_on_attribute(workers):

    # create remote object with children
    x = torch.Tensor([1, 2, 3])
    x = syft.LoggingTensor().on(x).send(workers["bob"])

    # call method on data tensor directly
    x.child.point_to_attr = "child.child"
    y = x.add(x)
    assert isinstance(y.get(), torch.Tensor)

    # call method on loggingtensor directly
    x.child.point_to_attr = "child"
    y = x.add(x)
    y = y.get()
    assert isinstance(y.child, syft.LoggingTensor)

    # # call method on zeroth attribute
    # x.child.point_to_attr = ""
    # y = x.add(x)
    # y = y.get()
    #
    # assert isinstance(y, torch.Tensor)
    # assert isinstance(y.child, syft.LoggingTensor)
    # assert isinstance(y.child.child, torch.Tensor)

    # call .get() on pinter to attribute (should error)
    x.child.point_to_attr = "child"
    try:
        x.get()
    except syft.exceptions.CannotRequestObjectAttribute as e:
        assert True


def test_grad_pointer(workers):
    """Tests the automatic creation of a .grad pointer when
    calling .send() on a tensor with requires_grad==True"""

    bob = workers["bob"]

    x = torch.tensor([1, 2, 3.0], requires_grad=True).send(bob)
    y = (x + x).sum()
    y.backward()

    assert (bob._objects[x.id_at_location].grad == torch.tensor([2, 2, 2.0])).all()


def test_move(workers):
    alice, bob = workers["alice"], workers["bob"]

    x = torch.tensor([1, 2, 3, 4, 5]).send(bob)

    assert x.id_at_location in bob._objects
    assert x.id_at_location not in alice._objects

    x.move(alice)

    assert x.id_at_location not in bob._objects
    assert x.id_at_location in alice._objects

    x = torch.tensor([1.0, 2, 3, 4, 5], requires_grad=True).send(bob)

    assert x.id_at_location in bob._objects
    assert x.id_at_location not in alice._objects

    x.move(alice)

    assert x.id_at_location not in bob._objects
    assert x.id_at_location in alice._objects

    alice.clear_objects()
    bob.clear_objects()
    x = torch.tensor([1.0, 2, 3, 4, 5]).send(bob)
    x.move(alice)

    assert len(alice._objects) == 1


def test_combine_pointers(workers):
    """
    Ensure that the sy.combine_pointers works as expected
    """

    bob = workers["bob"]
    alice = workers["alice"]

    x = th.tensor([1, 2, 3, 4, 5]).send(bob)
    y = th.tensor([1, 2, 3, 4, 5]).send(alice)

    a = x.combine(y)
    b = a + a

    c = b.get(sum_results=True)
    assert (c == th.tensor([4, 8, 12, 16, 20])).all()

    b = a + a
    c = b.get(sum_results=False)
    assert len(c) == 2
    assert (c[0] == th.tensor([2, 4, 6, 8, 10])).all


def test_remote_to_cpu_device(workers):
    """Ensure remote .to cpu works"""
    device = torch.device("cpu")
    bob = workers["bob"]

    x = th.tensor([1, 2, 3, 4, 5]).send(bob)
    x.to(device)


def test_get_remote_shape(workers):
    """Test pointer.shape functionality"""
    bob = workers["bob"]
    # tensor directly sent: shape stored at sending
    x = th.tensor([1, 2, 3, 4, 5]).send(bob)
    assert x.shape == torch.Size([5])
    # result of an operation: need to make a call to the remote worker
    y = x + x
    assert y.shape == torch.Size([5])


def test_remote_function_with_multi_ouput(workers):
    """
    Functions like .split return several tensors, registration and response
    must be made carefully in this case
    """
    bob = workers["bob"]

    tensor = torch.tensor([1, 2, 3, 4.0])
    ptr = tensor.send(bob)
    r_ptr = torch.split(ptr, 2)
    assert (r_ptr[0].get() == torch.tensor([1, 2.0])).all()

    tensor = torch.tensor([1, 2, 3, 4.0])
    ptr = tensor.send(bob)
    max_value, argmax_idx = torch.max(ptr, 0)

    assert max_value.get().item() == 4.0
    assert argmax_idx.get().item() == 3


def test_raising_error_when_item_func_called(workers):
    pointer = PointerTensor(id=1000, location=workers["alice"], owner=workers["me"])
    with pytest.raises(RuntimeError):
        pointer.item()


def test_fix_prec_on_pointer_tensor(workers):
    """
    Ensure .fix_precision() works as expected.
    """
    bob = workers["bob"]

    tensor = torch.tensor([1, 2, 3, 4.0])
    ptr = tensor.send(bob)

    ptr = ptr.fix_precision()
    remote_tensor = bob._objects[ptr.id_at_location]

    assert isinstance(ptr.child, PointerTensor)
    assert isinstance(remote_tensor.child, FixedPrecisionTensor)


def test_fix_prec_on_pointer_of_pointer(workers):
    """
    Ensure .fix_precision() works along a chain of pointers.
    """
    bob = workers["bob"]
    alice = workers["alice"]

    tensor = torch.tensor([1, 2, 3, 4.0])
    ptr = tensor.send(bob)
    ptr = ptr.send(alice)

    ptr = ptr.fix_precision()

    alice_tensor = alice._objects[ptr.id_at_location]
    remote_tensor = bob._objects[alice_tensor.id_at_location]

    assert isinstance(ptr.child, PointerTensor)
    assert isinstance(remote_tensor.child, FixedPrecisionTensor)


def test_float_prec_on_pointer_tensor(workers):
    """
    Ensure .float_precision() works as expected.
    """
    bob = workers["bob"]

    tensor = torch.tensor([1, 2, 3, 4.0])
    ptr = tensor.send(bob)
    ptr = ptr.fix_precision()

    ptr = ptr.float_precision()
    remote_tensor = bob._objects[ptr.id_at_location]

    assert isinstance(ptr.child, PointerTensor)
    assert isinstance(remote_tensor, torch.Tensor)


def test_float_prec_on_pointer_of_pointer(workers):
    """
    Ensure .float_precision() works along a chain of pointers.
    """
    bob = workers["bob"]
    alice = workers["alice"]

    tensor = torch.tensor([1, 2, 3, 4.0])
    ptr = tensor.send(bob)
    ptr = ptr.send(alice)
    ptr = ptr.fix_precision()

    ptr = ptr.float_precision()

    alice_tensor = alice._objects[ptr.id_at_location]
    remote_tensor = bob._objects[alice_tensor.id_at_location]

    assert isinstance(ptr.child, PointerTensor)
    assert isinstance(remote_tensor, torch.Tensor)



import torch
from typing import Callable
from syft.workers.abstract import AbstractWorker


def request_triple(
    crypto_provider: AbstractWorker,
    cmd: Callable,
    field: int,
    a_size: tuple,
    b_size: tuple,
    locations: list,
):
    """Generates a multiplication triple and sends it to all locations.

    Args:
        crypto_provider: worker you would like to request the triple from
        cmd: An equation in einsum notation.
        field: An integer representing the field size.
        a_size: A tuple which is the size that a should be.
        b_size: A tuple which is the size that b should be.
        locations: A list of workers where the triple should be shared between.

    Returns:
        A triple of AdditiveSharedTensors such that c_shared = cmd(a_shared, b_shared).
    """
    a = torch.randint(field, a_size)
    b = torch.randint(field, b_size)
    c = cmd(a, b)
    a_shared = a.share(*locations, field=field, crypto_provider=crypto_provider).child
    b_shared = b.share(*locations, field=field, crypto_provider=crypto_provider).child
    c_shared = c.share(*locations, field=field, crypto_provider=crypto_provider).child
    return a_shared, b_shared, c_shared



import pytest
import torch as th
import syft as sy

from syft.frameworks.torch.tensors.decorators.logging import LoggingTensor


def test_combine_pointers(workers):
    """
    Ensure that the sy.combine_pointers works as expected
    """

    bob = workers["bob"]
    alice = workers["alice"]

    x = th.tensor([1, 2, 3, 4, 5]).send(bob)
    y = th.tensor([1, 2, 3, 4, 5]).send(alice)

    a = sy.combine_pointers(*[x, y])
    b = a + a

    c = b.get(sum_results=True)
    assert (c == th.tensor([4, 8, 12, 16, 20])).all()

    b = a + a
    c = b.get(sum_results=False)
    assert len(c) == 2
    assert (c[0] == th.tensor([2, 4, 6, 8, 10])).all



import io
from os.path import exists, join
import time
from socket import gethostname
from OpenSSL import crypto, SSL
import pytest
import torch
import syft as sy
from syft.generic.frameworks.hook import hook_args
from syft.frameworks.torch.federated import utils

from syft.workers.websocket_client import WebsocketClientWorker
from syft.workers.websocket_server import WebsocketServerWorker

from test.conftest import instantiate_websocket_client_worker


PRINT_IN_UNITTESTS = False


@pytest.mark.parametrize("secure", [True, False])
def test_websocket_worker_basic(hook, start_proc, secure, tmpdir):
    """Evaluates that you can do basic tensor operations using
    WebsocketServerWorker in insecure and secure mode."""

    def create_self_signed_cert(cert_path, key_path):
        # create a key pair
        k = crypto.PKey()
        k.generate_key(crypto.TYPE_RSA, 1024)

        # create a self-signed cert
        cert = crypto.X509()
        cert.gmtime_adj_notBefore(0)
        cert.gmtime_adj_notAfter(1000)
        cert.set_pubkey(k)
        cert.sign(k, "sha1")

        # store keys and cert
        open(cert_path, "wb").write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))
        open(key_path, "wb").write(crypto.dump_privatekey(crypto.FILETYPE_PEM, k))

    kwargs = {
        "id": "secure_fed" if secure else "not_secure_fed",
        "host": "localhost",
        "port": 8766,
        "hook": hook,
    }

    if secure:
        # Create cert and keys
        cert_path = tmpdir.join("test.crt")
        key_path = tmpdir.join("test.key")
        create_self_signed_cert(cert_path, key_path)
        kwargs["cert_path"] = cert_path
        kwargs["key_path"] = key_path

    process_remote_worker = start_proc(WebsocketServerWorker, **kwargs)

    time.sleep(0.1)
    x = torch.ones(5)

    if secure:
        # unused args
        del kwargs["cert_path"]
        del kwargs["key_path"]

    kwargs["secure"] = secure
    remote_proxy = instantiate_websocket_client_worker(**kwargs)

    x = x.send(remote_proxy)
    y = x + x
    y = y.get()

    assert (y == torch.ones(5) * 2).all()

    del x

    remote_proxy.close()
    time.sleep(0.1)
    remote_proxy.remove_worker_from_local_worker_registry()
    process_remote_worker.terminate()


def test_websocket_workers_search(hook, start_remote_worker):
    """Evaluates that a client can search and find tensors that belong
    to another party"""
    # Args for initializing the websocket server and client
    server, remote_proxy = start_remote_worker(id="fed2", hook=hook, port=8767)

    # Sample tensor to store on the server
    sample_data = torch.tensor([1, 2, 3, 4]).tag("#sample_data", "#another_tag")
    _ = sample_data.send(remote_proxy)

    # Search for the tensor located on the server by using its tag
    results = remote_proxy.search("#sample_data", "#another_tag")

    assert results
    assert results[0].owner.id == "me"
    assert results[0].location.id == "fed2"

    # Search multiple times should still work
    results = remote_proxy.search("#sample_data", "#another_tag")

    assert results
    assert results[0].owner.id == "me"
    assert results[0].location.id == "fed2"

    remote_proxy.close()
    time.sleep(0.1)
    remote_proxy.remove_worker_from_local_worker_registry()
    server.terminate()


def test_list_objects_remote(hook, start_remote_worker):
    server, remote_proxy = start_remote_worker(id="fed-list-objects", hook=hook, port=8765)
    remote_proxy.clear_objects()

    x = torch.tensor([1, 2, 3]).send(remote_proxy)

    res = remote_proxy.list_objects_remote()

    x = torch.tensor([1, 2, 3]).send(remote_proxy)

    res = remote_proxy.list_objects_remote()
    res_dict = eval(res.replace("tensor", "torch.tensor"))
    assert len(res_dict) == 1

    y = torch.tensor([4, 5, 6]).send(remote_proxy)
    res = remote_proxy.list_objects_remote()
    res_dict = eval(res.replace("tensor", "torch.tensor"))
    assert len(res_dict) == 2

    # delete x before terminating the websocket connection
    del x
    del y
    time.sleep(0.1)
    remote_proxy.close()
    time.sleep(0.1)
    remote_proxy.remove_worker_from_local_worker_registry()
    server.terminate()


def test_objects_count_remote(hook, start_remote_worker):
    server, remote_proxy = start_remote_worker(id="fed-count-objects", hook=hook, port=8764)
    remote_proxy.clear_objects()

    x = torch.tensor([1, 2, 3]).send(remote_proxy)

    nr_objects = remote_proxy.objects_count_remote()
    assert nr_objects == 1

    y = torch.tensor([4, 5, 6]).send(remote_proxy)
    nr_objects = remote_proxy.objects_count_remote()
    assert nr_objects == 2

    x.get()
    nr_objects = remote_proxy.objects_count_remote()
    assert nr_objects == 1

    # delete remote object before terminating the websocket connection
    del y
    time.sleep(0.1)
    remote_proxy.close()
    time.sleep(0.1)
    remote_proxy.remove_worker_from_local_worker_registry()
    server.terminate()


def test_clear_objects_remote(hook, start_remote_worker):
    server, remote_proxy = start_remote_worker(id="fed-clear-objects", hook=hook, port=8769)

    x = torch.tensor([1, 2, 3]).send(remote_proxy, garbage_collect_data=False)
    y = torch.tensor(4).send(remote_proxy, garbage_collect_data=False)

    nr_objects = remote_proxy.objects_count_remote()
    assert nr_objects == 2

    remote_proxy.clear_objects_remote()
    nr_objects = remote_proxy.objects_count_remote()
    assert nr_objects == 0

    remote_proxy.close()
    remote_proxy.remove_worker_from_local_worker_registry()
    server.terminate()


def test_connect_close(hook, start_remote_worker):
    server, remote_proxy = start_remote_worker(id="fed-connect-close", hook=hook, port=8771)

    x = torch.tensor([1, 2, 3])
    x_ptr = x.send(remote_proxy)

    assert remote_proxy.objects_count_remote() == 1

    remote_proxy.close()

    time.sleep(0.1)

    remote_proxy.connect()

    assert remote_proxy.objects_count_remote() == 1

    x_val = x_ptr.get()
    assert (x_val == x).all()

    remote_proxy.close()
    remote_proxy.remove_worker_from_local_worker_registry()

    time.sleep(0.1)

    server.terminate()


def test_websocket_worker_multiple_output_response(hook, start_remote_worker):
    """Evaluates that you can do basic tensor operations using
    WebsocketServerWorker."""
    server, remote_proxy = start_remote_worker(id="socket_multiple_output", hook=hook, port=8768)

    x = torch.tensor([1.0, 3, 2])
    x = x.send(remote_proxy)

    p1, p2 = torch.sort(x)
    x1, x2 = p1.get(), p2.get()

    assert (x1 == torch.tensor([1.0, 2, 3])).all()
    assert (x2 == torch.tensor([0, 2, 1])).all()

    x.get()  # retrieve remote object before closing the websocket connection

    remote_proxy.close()
    server.terminate()


@pytest.mark.skipif(
    torch.__version__ >= "1.1",
    reason="bug in pytorch version 1.1.0, jit.trace returns raw C function",
)
def test_evaluate(hook, start_proc):  # pragma: no cover

    sy.local_worker.clear_objects()
    sy.generic.frameworks.hook.hook_args.hook_method_args_functions = {}
    sy.generic.frameworks.hook.hook_args.hook_method_response_functions = {}
    sy.generic.frameworks.hook.hook_args.get_tensor_type_functions = {}
    sy.generic.frameworks.hook.hook_args.register_response_functions = {}

    data, target = utils.iris_data_partial()

    dataset = sy.BaseDataset(data=data, targets=target)

    kwargs = {"id": "evaluate_remote", "host": "localhost", "port": 8790, "hook": hook}
    dataset_key = "iris"
    # TODO: check why unit test sometimes fails when WebsocketServerWorker is started from the unit test. Fails when run after test_federated_client.py
    # process_remote_worker = start_proc(WebsocketServerWorker, dataset=(dataset, dataset_key), verbose=True, **kwargs)

    remote_proxy = instantiate_websocket_client_worker(**kwargs)

    def loss_fn(pred, target):
        return torch.nn.functional.cross_entropy(input=pred, target=target)

    class Net(torch.nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.fc1 = torch.nn.Linear(4, 3)

            torch.nn.init.xavier_normal_(self.fc1.weight)

        def forward(self, x):
            x = torch.nn.functional.relu(self.fc1(x))
            return x

    model_untraced = Net()
    model = torch.jit.trace(model_untraced, data)
    loss_traced = torch.jit.trace(loss_fn, (torch.tensor([[0.3, 0.5, 0.2]]), torch.tensor([1])))

    pred = model(data)
    loss_before = loss_fn(target=target, pred=pred)
    if PRINT_IN_UNITTESTS:  # pragma: no cover
        print("Loss: {}".format(loss_before))

    # Create and send train config
    train_config = sy.TrainConfig(
        batch_size=4,
        model=model,
        loss_fn=loss_traced,
        model_id=None,
        loss_fn_id=None,
        optimizer_args=None,
        epochs=1,
    )
    train_config.send(remote_proxy)

    result = remote_proxy.evaluate(
        dataset_key=dataset_key, return_histograms=True, nr_bins=3, return_loss=True
    )

    len_dataset = result["nr_predictions"]
    hist_target = result["histogram_target"]

    if PRINT_IN_UNITTESTS:  # pragma: no cover
        print("Evaluation result before training: {}".format(result))

    assert len_dataset == 30
    assert (hist_target == [10, 10, 10]).all()

    remote_proxy.close()
    remote_proxy.remove_worker_from_local_worker_registry()
    # process_remote_worker.terminate()



import math
import logging

import torch
from torch.utils.data import Dataset

logger = logging.getLogger(__name__)


class BaseDataset:
    """
    This is a base class to be used for manipulating a dataset. This is composed
    of a .data attribute for inputs and a .targets one for labels. It is to
    be used like the MNIST Dataset object, and is useful to avoid handling
    the two inputs and label tensors separately.

    Args:

        data[list,torch tensors]: the data points
        targets: Corresponding labels of the data points
        transform: Function to transform the datapoints

    """

    def __init__(self, data, targets, transform=None):

        self.data = data
        self.targets = targets
        self.transform_ = transform

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):

        """
        Args:

            index[integer]: index of item to get

        Returns:

            data: Data points corresponding to the given index
            targets: Targets correspoding to given datapoint
        """
        data_elem = self.data[index]
        if self.transform_ is not None:
            # TODO: avoid passing through numpy domain
            data_elem = torch.tensor(self.transform_(data_elem.numpy()))

        return data_elem, self.targets[index]

    def transform(self, transform):

        """
         Allows a transform to be applied on given dataset.
         Args:

            transform: The transform to be applied on the data
        """

        # Transforms cannot be applied to Pointer, Fixed Precision or Float Precision tensors.
        if type(self.data) == torch.Tensor:

            self.data = transform(self.data)

        else:

            raise TypeError("Transforms can be applied only on torch tensors")

    def send(self, worker):
        """
        Args:

            worker[worker class]: worker to which the data must be sent

        Returns:

            self: Return the object instance with data sent to corresponding worker

        """

        self.data.send_(worker)
        self.targets.send_(worker)
        return self

    def get(self):
        """
        Gets the data back from respective workers.
        """

        self.data.get_()
        self.targets.get_()
        return self

    def fix_prec(self, *args, **kwargs):
        """
            Converts data of BaseDataset into fixed precision
        """
        self.data.fix_prec_(*args, **kwargs)
        self.targets.fix_prec_(*args, **kwargs)
        return self

    fix_precision = fix_prec

    def float_prec(self, *args, **kwargs):
        """
            Converts data of BaseDataset into float precision
        """
        self.data.float_prec_(*args, **kwargs)
        self.targets.float_prec_(*args, **kwargs)
        return self

    float_precision = float_prec

    def share(self, *args, **kwargs):
        """
            Share the data with the respective workers
        """
        self.data.share_(*args, **kwargs)
        self.targets.share_(*args, **kwargs)
        return self

    @property
    def location(self):
        """
            Get location of the data
        """
        return self.data.location


def dataset_federate(dataset, workers):
    """
    Add a method to easily transform a torch.Dataset or a sy.BaseDataset
    into a sy.FederatedDataset. The dataset given is split in len(workers)
    part and sent to each workers
    """
    logger.info("Scanning and sending data to {}...".format(", ".join([w.id for w in workers])))

    # take ceil to have exactly len(workers) sets after splitting
    data_size = math.ceil(len(dataset) / len(workers))

    # Fix for old versions of torchvision
    if not hasattr(dataset, "data"):
        if hasattr(dataset, "train_data"):
            dataset.data = dataset.train_data
        elif hasattr(dataset, "test_data"):
            dataset.data = dataset.test_data
        else:
            raise AttributeError("Could not find inputs in dataset")
    if not hasattr(dataset, "targets"):
        if hasattr(dataset, "train_labels"):
            dataset.targets = dataset.train_labels
        elif hasattr(dataset, "test_labels"):
            dataset.targets = dataset.test_labels
        else:
            raise AttributeError("Could not find targets in dataset")

    datasets = []
    data_loader = torch.utils.data.DataLoader(dataset, batch_size=data_size)
    for dataset_idx, (data, targets) in enumerate(data_loader):
        worker = workers[dataset_idx % len(workers)]
        logger.debug("Sending data to worker %s", worker.id)
        data = data.send(worker)
        targets = targets.send(worker)
        datasets.append(BaseDataset(data, targets))  # .send(worker)

    logger.debug("Done!")
    return FederatedDataset(datasets)


Dataset.federate = dataset_federate
BaseDataset.federate = dataset_federate


class FederatedDataset:
    def __init__(self, datasets):
        """This class takes a list of datasets, each of which is supposed
        to be already sent to a remote worker (they have a location), and
        acts like a dictionary based on the worker ids.
        It serves like an input for the FederatedDataLoader.
        Args:
            datasets (list): list of remote Datasets
        """
        self.datasets = {}
        for dataset in datasets:
            worker_id = dataset.data.location.id
            self.datasets[worker_id] = dataset

        # Check that data and targets for a worker are consistent
        for worker_id in self.workers:
            dataset = self.datasets[worker_id]
            assert len(dataset.data) == len(
                dataset.targets
            ), "On each worker, the input and target must have the same number of rows."

    @property
    def workers(self):
        """
           Returns: list of workers
        """

        return list(self.datasets.keys())

    def __getitem__(self, worker_id):
        """
           Args:
                   worker_id[str,int]: ID of respective worker

           Returns: Get Datasets from the respective worker
        """

        return self.datasets[worker_id]

    def __len__(self):

        return sum([len(dataset) for w, dataset in self.datasets.items()])

    def __repr__(self):

        fmt_str = "FederatedDataset\n"
        fmt_str += "    Distributed accross: {}\n".format(", ".join(str(x) for x in self.workers))
        fmt_str += "    Number of datapoints: {}\n".format(self.__len__())
        return fmt_str



import pytest
import torch
import syft as sy
from syft.frameworks.torch.linalg import inv_sym
from test.efficiency_tests.assertions import assert_time


@assert_time(max_time=40)
def test_inv_sym(hook, workers):
    torch.manual_seed(42)  # Truncation might not always work so we set the random seed
    N = 100
    K = 2
    bob = workers["bob"]
    alice = workers["alice"]
    crypto_prov = sy.VirtualWorker(hook, id="crypto_prov")

    x = torch.randn(N, K).fix_precision().share(bob, alice, crypto_provider=crypto_prov)
    gram = x.t().matmul(x)
    gram_inv = inv_sym(gram)



import torch


def test_section_1_differential_privacy():
    """This tests the Udacity course content found at
    https://github.com/Udacity/private-ai
    """

    # the number of entries in our database
    num_entries = 5000

    db = torch.rand(num_entries) > 0.5

    db = torch.rand(num_entries) > 0.5

    def get_parallel_db(db, remove_index):
        return torch.cat((db[0:remove_index], db[remove_index + 1 :]))

    get_parallel_db(db, 52352)

    def get_parallel_dbs(db):
        parallel_dbs = list()

        for i in range(len(db)):
            pdb = get_parallel_db(db, i)
            parallel_dbs.append(pdb)

        return parallel_dbs

    pdbs = get_parallel_dbs(db)

    def create_db_and_parallels(num_entries):
        db = torch.rand(num_entries) > 0.5
        pdbs = get_parallel_dbs(db)

        return db, pdbs

    db, pdbs = create_db_and_parallels(20)

    db, pdbs = create_db_and_parallels(5000)

    def query(db):
        return db.sum()

    full_db_result = query(db)

    sensitivity = 0
    for pdb in pdbs:
        pdb_result = query(pdb)

        db_distance = torch.abs(pdb_result - full_db_result)

        if db_distance > sensitivity:
            sensitivity = db_distance

    def sensitivity(query, n_entries=1000):

        db, pdbs = create_db_and_parallels(n_entries)

        full_db_result = query(db)

        max_distance = 0
        for pdb in pdbs:
            pdb_result = query(pdb)

            db_distance = torch.abs(pdb_result - full_db_result)

            if db_distance > max_distance:
                max_distance = db_distance

        return max_distance

    def query(db):
        return db.float().mean()

    sensitivity(query)

    db, pdbs = create_db_and_parallels(20)

    db

    def query(db, threshold=5):
        return (db.sum() > threshold).float()

    for i in range(10):
        sens_f = sensitivity(query, n_entries=10)
        print(sens_f)

    db, _ = create_db_and_parallels(100)

    pdb = get_parallel_db(db, remove_index=10)

    db[10]

    sum(db)

    # differencing attack using sum query

    sum(db) - sum(pdb)

    # differencing attack using mean query

    (sum(db).float() / len(db)) - (sum(pdb).float() / len(pdb))

    # differencing attack using threshold

    (sum(db).float() > 49) - (sum(pdb).float() > 49)

    def query(db):

        true_result = torch.mean(db.float())

        first_coin_flip = (torch.rand(len(db)) > 0.5).float()
        second_coin_flip = (torch.rand(len(db)) > 0.5).float()

        augmented_database = db.float() * first_coin_flip + (1 - first_coin_flip) * second_coin_flip

        db_result = torch.mean(augmented_database.float()) * 2 - 0.5

        return db_result, true_result

    db, pdbs = create_db_and_parallels(10)
    private_result, true_result = query(db)
    print("With Noise:" + str(private_result))
    print("Without Noise:" + str(true_result))

    db, pdbs = create_db_and_parallels(100)
    private_result, true_result = query(db)
    print("With Noise:" + str(private_result))
    print("Without Noise:" + str(true_result))

    db, pdbs = create_db_and_parallels(1000)
    private_result, true_result = query(db)
    print("With Noise:" + str(private_result))
    print("Without Noise:" + str(true_result))

    db, pdbs = create_db_and_parallels(10000)
    private_result, true_result = query(db)
    print("With Noise:" + str(private_result))
    print("Without Noise:" + str(true_result))

    def query(db, noise=0.2):

        true_result = torch.mean(db.float())

        first_coin_flip = (torch.rand(len(db)) > noise).float()
        second_coin_flip = (torch.rand(len(db)) > 0.5).float()

        augmented_database = db.float() * first_coin_flip + (1 - first_coin_flip) * second_coin_flip

        sk_result = augmented_database.float().mean()

        private_result = ((sk_result / noise) - 0.5) * noise / (1 - noise)

        return private_result, true_result

    db, pdbs = create_db_and_parallels(100)
    private_result, true_result = query(db, noise=0.1)
    print("With Noise:" + str(private_result))
    print("Without Noise:" + str(true_result))

    db, pdbs = create_db_and_parallels(100)
    private_result, true_result = query(db, noise=0.2)
    print("With Noise:" + str(private_result))
    print("Without Noise:" + str(true_result))

    db, pdbs = create_db_and_parallels(100)
    private_result, true_result = query(db, noise=0.4)
    print("With Noise:" + str(private_result))
    print("Without Noise:" + str(true_result))

    db, pdbs = create_db_and_parallels(100)
    private_result, true_result = query(db, noise=0.8)
    print("With Noise:" + str(private_result))
    print("Without Noise:" + str(true_result))

    db, pdbs = create_db_and_parallels(10000)
    private_result, true_result = query(db, noise=0.8)
    print("With Noise:" + str(private_result))
    print("Without Noise:" + str(true_result))

    db, pdbs = create_db_and_parallels(100)

    def query(db):
        return torch.sum(db.float())

    # def M(db):
    #     query(db)  # + noise
    #
    query(db)

    epsilon = 0.0001

    import numpy as np

    db, pdbs = create_db_and_parallels(100)

    def sum_query(db):
        return db.sum()

    def laplacian_mechanism(db, query, sensitivity):

        beta = sensitivity / epsilon
        noise = torch.tensor(np.random.laplace(0, beta, 1))

        return query(db) + noise

    def mean_query(db):
        return torch.mean(db.float())

    laplacian_mechanism(db, sum_query, 1)

    laplacian_mechanism(db, mean_query, 1 / 100)

    import numpy as np

    num_teachers = 10  # we're working with 10 partner hospitals
    num_examples = 10000  # the size of OUR dataset
    num_labels = 10  # number of lablels for our classifier

    preds = (
        (np.random.rand(num_teachers, num_examples) * num_labels).astype(int).transpose(1, 0)
    )  # fake predictions

    new_labels = list()
    for an_image in preds:

        label_counts = np.bincount(an_image, minlength=num_labels)

        epsilon = 0.1
        beta = 1 / epsilon

        for i in range(len(label_counts)):
            label_counts[i] += np.random.laplace(0, beta, 1)

        new_label = np.argmax(label_counts)

        new_labels.append(new_label)

    labels = np.array([9, 9, 3, 6, 9, 9, 9, 9, 8, 2])
    counts = np.bincount(labels, minlength=10)
    query_result = np.argmax(counts)
    query_result

    from syft.frameworks.torch.differential_privacy import pate

    num_teachers, num_examples, num_labels = (100, 100, 10)
    preds = (np.random.rand(num_teachers, num_examples) * num_labels).astype(int)  # fake preds
    indices = (np.random.rand(num_examples) * num_labels).astype(int)  # true answers

    preds[:, 0:10] *= 0

    data_dep_eps, data_ind_eps = pate.perform_analysis(
        teacher_preds=preds, indices=indices, noise_eps=0.1, delta=1e-5
    )

    assert data_dep_eps < data_ind_eps

    data_dep_eps, data_ind_eps = pate.perform_analysis(
        teacher_preds=preds, indices=indices, noise_eps=0.1, delta=1e-5
    )
    print("Data Independent Epsilon:", data_ind_eps)
    print("Data Dependent Epsilon:", data_dep_eps)

    preds[:, 0:50] *= 0

    data_dep_eps, data_ind_eps = pate.perform_analysis(
        teacher_preds=preds, indices=indices, noise_eps=0.1, delta=1e-5, moments=20
    )
    print("Data Independent Epsilon:", data_ind_eps)
    print("Data Dependent Epsilon:", data_dep_eps)


def test_section_2_federated_learning(hook):
    """This tests the Udacity course content found at
    https://github.com/Udacity/private-ai
    """

    import torch as th

    x = th.tensor([1, 2, 3, 4, 5])
    x

    y = x + x

    print(y)

    import syft as sy

    # commented out because the test needs to use the global one
    # hook = sy.TorchHook(th)

    th.tensor([1, 2, 3, 4, 5])

    bob = sy.VirtualWorker(hook, id="bob_udacity")

    bob._objects

    x = th.tensor([1, 2, 3, 4, 5])

    x = x.send(bob)

    bob._objects

    assert len(bob._objects) == 1

    x.location

    x.id_at_location

    x.id

    x.owner

    hook.local_worker

    x

    x = x.get()
    x

    bob._objects

    assert len(bob._objects) == 0

    alice = sy.VirtualWorker(hook, id="alice_udacity")

    x = th.tensor([1, 2, 3, 4, 5])

    x_ptr = x.send(bob, alice)

    x_ptr.get()

    x = th.tensor([1, 2, 3, 4, 5]).send(bob, alice)

    x.get(sum_results=True)

    x = th.tensor([1, 2, 3, 4, 5]).send(bob)
    y = th.tensor([1, 1, 1, 1, 1]).send(bob)

    x
    y

    z = x + y

    z

    z = z.get()
    z

    z = th.add(x, y)
    z

    z = z.get()
    z

    x = th.tensor([1.0, 2, 3, 4, 5], requires_grad=True).send(bob)
    y = th.tensor([1.0, 1, 1, 1, 1], requires_grad=True).send(bob)

    z = (x + y).sum()

    z.backward()

    x = x.get()

    x

    x.grad

    input = th.tensor([[1.0, 1], [0, 1], [1, 0], [0, 0]], requires_grad=True).send(bob)
    target = th.tensor([[1.0], [1], [0], [0]], requires_grad=True).send(bob)

    weights = th.tensor([[0.0], [0.0]], requires_grad=True).send(bob)

    for i in range(10):
        pred = input.mm(weights)

        loss = ((pred - target) ** 2).sum()

        loss.backward()

        weights.data.sub_(weights.grad * 0.1)
        weights.grad *= 0

        print(loss.get().data)

    bob = bob.clear_objects()

    assert len(bob._objects) == 0

    x = th.tensor([1, 2, 3, 4, 5]).send(bob)

    assert len(bob._objects) == 1

    del x

    assert len(bob._objects) == 0

    x = th.tensor([1, 2, 3, 4, 5]).send(bob)

    assert len(bob._objects) == 1

    x = "asdf"

    assert len(bob._objects) == 0

    x = th.tensor([1, 2, 3, 4, 5]).send(bob)

    bob._objects

    x = "asdf"

    bob._objects

    del x

    bob._objects

    bob = bob.clear_objects()
    bob._objects

    for i in range(1000):
        x = th.tensor([1, 2, 3, 4, 5]).send(bob)

    assert len(bob._objects) == 1

    x = th.tensor([1, 2, 3, 4, 5]).send(bob)
    y = th.tensor([1, 1, 1, 1, 1])

    # throws error
    # z = x + y

    x = th.tensor([1, 2, 3, 4, 5]).send(bob)
    y = th.tensor([1, 1, 1, 1, 1]).send(alice)

    # throws error
    # z = x + y

    from torch import nn, optim

    # A Toy Dataset
    data = th.tensor([[1.0, 1], [0, 1], [1, 0], [0, 0]], requires_grad=True)
    target = th.tensor([[1.0], [1], [0], [0]], requires_grad=True)

    # A Toy Model
    model = nn.Linear(2, 1)

    opt = optim.SGD(params=model.parameters(), lr=0.1)

    def train(iterations=20):
        for iter in range(iterations):
            opt.zero_grad()

            pred = model(data)

            loss = ((pred - target) ** 2).sum()

            loss.backward()

            opt.step()

            print(loss.data)

    train()

    data_bob = data[0:2].send(bob)
    target_bob = target[0:2].send(bob)

    data_alice = data[2:4].send(alice)
    target_alice = target[2:4].send(alice)

    datasets = [(data_bob, target_bob), (data_alice, target_alice)]

    def train(iterations=20):

        model = nn.Linear(2, 1)
        opt = optim.SGD(params=model.parameters(), lr=0.1)

        for iter in range(iterations):

            for _data, _target in datasets:
                # send model to the data
                model = model.send(_data.location)

                # do normal training
                opt.zero_grad()
                pred = model(_data)
                loss = ((pred - _target) ** 2).sum()
                loss.backward()
                opt.step()

                # get smarter model back
                model = model.get()

                print(loss.get())

    train()

    bob.clear_objects()
    alice.clear_objects()

    x = th.tensor([1, 2, 3, 4, 5]).send(bob)

    x = x.send(alice)

    bob._objects

    alice._objects

    y = x + x

    y

    bob._objects

    alice._objects

    jon = sy.VirtualWorker(hook, id="jon")

    bob.clear_objects()
    alice.clear_objects()

    x = th.tensor([1, 2, 3, 4, 5]).send(bob).send(alice)

    bob._objects

    alice._objects

    x = x.get()
    x

    bob._objects

    alice._objects

    x = x.get()
    x

    bob._objects

    bob.clear_objects()
    alice.clear_objects()

    x = th.tensor([1, 2, 3, 4, 5]).send(bob).send(alice)

    bob._objects

    alice._objects

    del x

    bob._objects

    alice._objects

    bob.clear_objects()
    alice.clear_objects()

    x = th.tensor([1, 2, 3, 4, 5]).send(bob)

    bob._objects

    alice._objects

    x.move(alice)

    bob._objects

    alice._objects

    x = th.tensor([1, 2, 3, 4, 5]).send(bob).send(alice)

    bob._objects

    alice._objects

    x.remote_get()

    bob._objects

    alice._objects

    x.move(bob)

    x

    bob._objects

    alice._objects


def test_section_3_securing_fl(hook):
    """This tests the Udacity course content found at
    https://github.com/Udacity/private-ai
    """

    import syft as sy
    import torch as th

    # hook = sy.TorchHook(th)
    from torch import nn, optim

    # create a couple workers

    bob = sy.VirtualWorker(hook, id="bob_udacity_3")
    alice = sy.VirtualWorker(hook, id="alice_udacity_3")
    secure_worker = sy.VirtualWorker(hook, id="secure_worker_udacity_3")

    bob.add_workers([alice, secure_worker])
    alice.add_workers([bob, secure_worker])
    secure_worker.add_workers([alice, bob])

    # A Toy Dataset
    data = th.tensor([[0, 0], [0, 1], [1, 0], [1, 1.0]], requires_grad=True)
    target = th.tensor([[0], [0], [1], [1.0]], requires_grad=True)

    # get pointers to training data on each worker by
    # sending some training data to bob and alice
    bobs_data = data[0:2].send(bob)
    bobs_target = target[0:2].send(bob)

    alices_data = data[2:].send(alice)
    alices_target = target[2:].send(alice)

    # Iniitalize A Toy Model
    model = nn.Linear(2, 1)

    bobs_model = model.copy().send(bob)
    alices_model = model.copy().send(alice)

    bobs_opt = optim.SGD(params=bobs_model.parameters(), lr=0.1)
    alices_opt = optim.SGD(params=alices_model.parameters(), lr=0.1)

    for i in range(10):
        # Train Bob's Model
        bobs_opt.zero_grad()
        bobs_pred = bobs_model(bobs_data)
        bobs_loss = ((bobs_pred - bobs_target) ** 2).sum()
        bobs_loss.backward()

        bobs_opt.step()
        bobs_loss = bobs_loss.get().data

        # Train Alice's Model
        alices_opt.zero_grad()
        alices_pred = alices_model(alices_data)
        alices_loss = ((alices_pred - alices_target) ** 2).sum()
        alices_loss.backward()

        alices_opt.step()
        alices_loss = alices_loss.get().data
        alices_loss

    alices_model.move(secure_worker)
    bobs_model.move(secure_worker)

    with th.no_grad():

        model.weight.set_(((alices_model.weight.data + bobs_model.weight.data) / 2).get())
        model.bias.set_(((alices_model.bias.data + bobs_model.bias.data) / 2).get())

    iterations = 10
    worker_iters = 5

    for a_iter in range(iterations):

        bobs_model = model.copy().send(bob)
        alices_model = model.copy().send(alice)

        bobs_opt = optim.SGD(params=bobs_model.parameters(), lr=0.1)
        alices_opt = optim.SGD(params=alices_model.parameters(), lr=0.1)

        for wi in range(worker_iters):
            # Train Bob's Model
            bobs_opt.zero_grad()
            bobs_pred = bobs_model(bobs_data)
            bobs_loss = ((bobs_pred - bobs_target) ** 2).sum()
            bobs_loss.backward()

            bobs_opt.step()
            bobs_loss = bobs_loss.get().data

            # Train Alice's Model
            alices_opt.zero_grad()
            alices_pred = alices_model(alices_data)
            alices_loss = ((alices_pred - alices_target) ** 2).sum()
            alices_loss.backward()

            alices_opt.step()
            alices_loss = alices_loss.get().data

        alices_model.move(secure_worker)
        bobs_model.move(secure_worker)

        with th.no_grad():

            model.weight.set_(((alices_model.weight.data + bobs_model.weight.data) / 2).get())
            model.bias.set_(((alices_model.bias.data + bobs_model.bias.data) / 2).get())

        print("Bob:" + str(bobs_loss) + " Alice:" + str(alices_loss))

    preds = model(data)
    loss = ((preds - target) ** 2).sum()

    print(preds)
    print(target)
    print(loss.data)

    x = 5

    bob_x_share = 2
    alice_x_share = 3

    decrypted_x = bob_x_share + alice_x_share
    decrypted_x

    bob_x_share = 2 * 2
    alice_x_share = 3 * 2

    decrypted_x = bob_x_share + alice_x_share
    decrypted_x

    # encrypted "5"
    bob_x_share = 2
    alice_x_share = 3

    # encrypted "7"
    bob_y_share = 5
    alice_y_share = 2

    # encrypted 5 + 7
    bob_z_share = bob_x_share + bob_y_share
    alice_z_share = alice_x_share + alice_y_share

    decrypted_z = bob_z_share + alice_z_share
    decrypted_z

    x = 5

    Q = 23740629843760239486723

    bob_x_share = 23552870267  # <- a random number
    alice_x_share = Q - bob_x_share + x
    alice_x_share

    (bob_x_share + alice_x_share) % Q

    x_share = (2, 5, 7)

    import random

    Q = 23740629843760239486723

    def encrypt(x, n_share=3):

        shares = list()

        for i in range(n_share - 1):
            shares.append(random.randint(0, Q))

        shares.append(Q - (sum(shares) % Q) + x)

        return tuple(shares)

    def decrypt(shares):
        return sum(shares) % Q

    shares = encrypt(3)
    shares

    decrypt(shares)

    def add(a, b):
        c = list()
        for i in range(len(a)):
            c.append((a[i] + b[i]) % Q)
        return tuple(c)

    x = encrypt(5)
    y = encrypt(7)
    z = add(x, y)
    decrypt(z)

    BASE = 10
    PRECISION = 4

    def encode(x):
        return int((x * (BASE ** PRECISION)) % Q)

    def decode(x):
        return (x if x <= Q / 2 else x - Q) / BASE ** PRECISION

    encode(3.5)

    decode(35000)

    x = encrypt(encode(5.5))
    y = encrypt(encode(2.3))
    z = add(x, y)
    decode(decrypt(z))

    bob = bob.clear_objects()
    alice = alice.clear_objects()
    secure_worker = secure_worker.clear_objects()

    x = th.tensor([1, 2, 3, 4, 5])

    x = x.share(bob, alice, secure_worker)

    bob._objects

    y = x + x

    y

    y.get()

    x = th.tensor([0.1, 0.2, 0.3])

    x = x.fix_prec()

    x.child.child

    y = x + x

    y = y.float_prec()
    y

    x = th.tensor([0.1, 0.2, 0.3])

    x = x.fix_prec().share(bob, alice, secure_worker)

    y = x + x

    y.get().float_prec()



from syft.generic.frameworks.hook import hook_args
from syft.generic.pointers.pointer_tensor import PointerTensor
import torch
import numpy as np


def test_build_rule_syft_tensors_and_pointers():
    pointer = PointerTensor(id=1000, location="location", owner="owner", garbage_collect_data=False)
    result = hook_args.build_rule(([torch.tensor([1, 2]), pointer], 42))
    assert result == ([1, 1], 0)


def test_build_rule_numpy():
    arr = np.array([2.0, 3.0, 4.0])
    result = hook_args.build_rule([arr, arr + 2, [2, 4, "string"]])
    assert result == [1, 1, [0, 0, 0]]


def test_backward_multiple_use(workers):
    """
    Test using backward() in different contexts (FL or Encrypted) within
    the same session.
    """
    big_hospital, small_hospital, crypto_provider = (
        workers["bob"],
        workers["alice"],
        workers["james"],
    )

    # A Toy Model
    class Net(torch.nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.fc = torch.nn.Linear(2, 1)

        def forward(self, x):
            x = self.fc(x)
            return x

    def federated():
        # A Toy Dataset
        data = torch.tensor([[0, 0], [0, 1], [1, 0], [1, 1.0]])
        target = torch.tensor([[0], [0], [1], [1.0]])

        model = Net()

        model_weight = model.fc.weight.copy()

        # Training Logic
        opt = torch.optim.SGD(params=model.parameters(), lr=0.1)

        data = data.send(big_hospital)
        target = target.send(big_hospital)

        # NEW) send model to correct worker
        model.send(data.location)

        # 1) erase previous gradients (if they exist)
        opt.zero_grad()

        # 2) make a prediction
        pred = model(data)

        # 3) calculate how much we missed
        loss = ((pred - target) ** 2).sum()

        # 4) figure out which weights caused us to miss
        loss.backward()

        # 5) change those weights
        opt.step()

        assert (model_weight - model.get().fc.weight).sum().abs() > 1.0e-3

    def encrypted():
        # A Toy Dataset
        data2 = torch.tensor([[0, 0], [0, 1], [1, 0], [1, 1.0]])
        target2 = torch.tensor([[0], [0], [1], [1.0]])

        model2 = Net()

        model2_weight = model2.fc.weight.copy()

        # We encode everything
        data2 = data2.fix_precision().share(
            big_hospital, small_hospital, crypto_provider=crypto_provider, requires_grad=True
        )
        target2 = target2.fix_precision().share(
            big_hospital, small_hospital, crypto_provider=crypto_provider, requires_grad=True
        )
        model2 = model2.fix_precision().share(
            big_hospital, small_hospital, crypto_provider=crypto_provider, requires_grad=True
        )

        opt2 = torch.optim.SGD(params=model2.parameters(), lr=0.1).fix_precision()

        # 1) erase previous gradients (if they exist)
        opt2.zero_grad()

        # 2) make a prediction
        pred2 = model2(data2)

        # 3) calculate how much we missed
        loss2 = ((pred2 - target2) ** 2).sum()

        # 4) figure out which weights caused us to miss
        loss2.backward()

        # 5) change those weights
        opt2.step()

        weight_diff = (model2_weight - model2.fc.weight.get().float_prec()).sum().abs()
        assert weight_diff > 1.0e-3

    federated()
    encrypted()



import syft

if syft.dependency_check.tfe_available:
    from syft.frameworks.keras import layers
    from syft.frameworks.keras import model
    from syft.frameworks.keras.hook import KerasHook

    __all__ = ["KerasHook"]
